from importlib.resources import files
from lxml import etree
from optparse import OptionParser

import logging
import sys


try:
    stdout = sys.stdout.buffer
except AttributeError:
    # This can happen in (doc-)tests e. g. in plone.app.testing:
    stdout = None

strparam = etree.XSLT.strparam

logger = logging.getLogger("diazo")

namespaces = dict(
    diazo="http://namespaces.plone.org/diazo",
    css="http://namespaces.plone.org/diazo/css",
    old1="http://openplans.org/deliverance",
    old2="http://namespaces.plone.org/xdv",
    oldcss1="http://namespaces.plone.org/xdv+css",
    oldcss2="http://namespaces.plone.org/diazo+css",
    xml="http://www.w3.org/XML/1998/namespace",
    xsl="http://www.w3.org/1999/XSL/Transform",
)


def localname(name):
    return name.rsplit("}", 1)[1]


def namespace(name):
    return name.rsplit("}", 1)[0][1:]


def fullname(namespace, name):
    return f"{{{namespace}}}{name}"  # NOQA: S001


AC_READ_FILE = etree.XSLTAccessControl(
    read_file=True,
    write_file=False,
    create_dir=False,
    read_network=False,
    write_network=False,
)
AC_READ_NET = etree.XSLTAccessControl(
    read_file=True,
    write_file=False,
    create_dir=False,
    read_network=True,
    write_network=False,
)


class CustomResolver(etree.Resolver):
    def __init__(self, data):
        self.data = data

    def resolve(self, url, pubid, context):
        output = self.data.get(url)
        if output is not None:
            return self.resolve_string(output, context)


class LoggingXSLTWrapper:
    def __init__(self, xslt, logger):
        self.xslt = xslt
        self.logger = logger

    def __call__(self, *args, **kw):
        result = self.xslt(*args, **kw)
        for msg in self.xslt.error_log:
            if msg.type == etree.ErrorTypes.ERR_OK:
                self.logger.debug(msg.message)
            else:
                self.logger.debug(msg)
        return result


def pkg_parse(name, parser=None):
    path = files("diazo") / name
    with path.open() as f:
        return etree.parse(f, parser=parser)


def pkg_xsl(name, parser=None):
    return LoggingXSLTWrapper(etree.XSLT(pkg_parse(name, parser)), logger)


def quote_param(value):
    """Quote for passing as an XSL parameter.

    Works with strings, booleans, numbers and None.
    """

    if isinstance(value, str):
        return strparam(value)
    elif isinstance(value, bool):
        return value and "true()" or "false()"
    elif isinstance(value, (int,) + (float,)):
        value = repr(value)
    elif value is None:
        return "/.."
    else:
        raise ValueError("Cannot convert %s", value)


def split_params(s):
    """Turn foo,bar=baz into {'foo': None, 'bar': 'baz'}"""

    xsl_params = {}
    for param in s.split(","):
        tokens = [t.strip() for t in param.split("=")]
        xsl_params[tokens[0]] = len(tokens) > 1 and tokens[1] or None
    return xsl_params


def _createOptionParser(usage):
    parser = OptionParser(usage=usage)
    parser.add_option(
        "-o",
        "--output",
        metavar="output.xsl",
        help="Output filename (instead of stdout)",
        dest="output",
        default=stdout,
    )
    parser.add_option(
        "-p",
        "--pretty-print",
        action="store_true",
        help="Pretty print output (may alter rendering in browser)",
        dest="pretty_print",
        default=False,
    )
    parser.add_option(
        "--trace",
        action="store_true",
        help="Compiler trace logging",
        dest="trace",
        default=False,
    )
    parser.add_option(
        "-a",
        "--absolute-prefix",
        metavar="/",
        help="relative urls in the theme file will be made into absolute "
        "links with this prefix.",
        dest="absolute_prefix",
        default=None,
    )
    parser.add_option(
        "-i",
        "--includemode",
        metavar="INC",
        help="include mode (document, ssi, ssiwait or esi)",
        dest="includemode",
        default=None,
    )
    parser.add_option(
        "-n",
        "--network",
        action="store_true",
        help="Allow reads to the network to fetch resources",
        dest="read_network",
        default=False,
    )
    parser.add_option(
        "-t",
        "--theme",
        metavar="theme.html",
        help="Theme file",
        dest="theme",
        default=None,
    )
    parser.add_option(
        "-r",
        "--rules",
        metavar="rules.xml",
        help="Diazo rules file",
        dest="rules",
        default=None,
    )
    parser.add_option(
        "-c",
        "--custom-parameters",
        metavar="param1,param2=defaultval",
        help="Comma-separated list of custom parameter names with optional "
        "default values that the compiled theme will be able accept "
        "when run",
        dest="xsl_params",
        default=None,
    )
    parser.add_option(
        "-e",
        "--extra",
        metavar="extra.xsl",
        help="Extra XSL to be included in the transform "
        "(deprecated, use inline xsl in the rules instead)",
        dest="extra",
        default=None,
    )
    return parser



#!/usr/bin/env python

from diazo.rules import process_rules
from diazo.utils import pkg_xsl
from lxml import etree

import logging


logger = logging.getLogger("diazo")

_runtrace_to_html = pkg_xsl("runtrace_to_html.xsl")


def log_to_xml_string(error_log):
    msgs = [log.message for log in error_log if log.message.startswith("<runtrace ")]
    return """
<runtrace xmlns:css="http://namespaces.plone.org/diazo/css">
    {message:s}
</runtrace>
""".format(
        message="".join(msgs)
    )


def generate_runtrace(rules, error_log, rules_parser=None):
    """Annotate a rules file with the results of a transformation"""

    def condition_name(trace):
        """Generate attribute name for this entry"""
        for k in trace.attrib.keys():
            if k == "theme_xmlid":
                continue
            if k.startswith("{http://namespaces.plone.org/diazo/css}"):
                continue
            return "runtrace-" + k

    rules_doc = process_rules(
        rules,
        rules_parser=rules_parser,
        stop="add_identifiers",
    )
    trace_doc = etree.XML(log_to_xml_string(error_log))

    for trace in trace_doc.xpath("/runtrace/runtrace"):
        for el in rules_doc.xpath("id('" + trace.attrib["theme_xmlid"] + "')"):
            el.set(condition_name(trace), trace.text or "")
    return rules_doc


def runtrace_to_html(runtrace_doc):
    """Convert the runtrace document into HTML"""
    return _runtrace_to_html(runtrace_doc)


def error_log_to_html(error_log):
    """Convert an error log into an HTML representation"""
    doc = etree.Element("ul")
    for log in error_log:
        if log.message.startswith("<runtrace "):
            continue
        el = etree.Element("li")
        el.attrib["class"] = (
            "domain_{domain_name} level_{level_name} type_{type_name}".format(  # NOQA: E501
                domain_name=log.domain_name,
                level_name=log.level_name,
                type_name=log.type_name,
            )
        )
        el.text = "{msg:s} [{line:d}:{column:d}]".format(
            msg=log.message,
            line=log.line,
            column=log.column,
        )
        doc.append(el)
    return doc


def generate_debug_html(
    base_url,
    rules=None,
    error_log=None,
    rules_parser=None,
):
    """Generate an HTML node with debug info"""

    def newElement(tag, content, **kwargs):
        n = etree.Element(tag, **kwargs)
        if getattr(content, "tag", None):
            n.append(content)
        else:
            n.text = content
        return n

    debug_output_iframe = etree.Element("div", id="diazo-debug-iframe")
    debug_output_iframe.attrib["style"] = "display:none"
    debug_output_iframe.attrib["data-iframe"] = "diazo-debug"
    debug_output_iframe.attrib["data-iframe-style"] = ""
    debug_output_iframe.attrib["data-iframe-position"] = "bottom"
    debug_output_iframe.attrib["data-iframe-resources"] = (
        base_url
        + "/diazo-debug.css;"
        + base_url
        + "/jquery-1.8.3.min.js;"
        + base_url
        + "/diazo-debug.js"
    )
    debug_output = etree.Element("div", id="diazo-debug")

    if error_log:
        debug_output.append(
            newElement(
                "section",
                error_log_to_html(error_log),
                id="diazo_error_log",
            ),
        )

    try:
        runtrace_doc = generate_runtrace(rules, error_log, rules_parser)
        debug_output.append(
            newElement(
                "section",
                runtrace_to_html(runtrace_doc).getroot(),
                id="diazo_runtrace",
            ),
        )
    except etree.XMLSyntaxError:
        debug_output.append(
            newElement(
                "section",
                "Rules document could not be parsed!",
                id="diazo_runtrace",
            ),
        )

    debug_output_iframe.append(debug_output)
    debug_output_iframe.append(
        newElement(
            "script",
            " ",
            text="text/javascript",
            src=base_url + "/iframe.js",
        ),
    )
    return debug_output_iframe



#!/usr/bin/env python
"""\
Usage: %prog [-r] RULES

Preprocess RULES, an diazo rules file
"""

from diazo.cssrules import convert_css_selectors
from diazo.utils import _createOptionParser
from diazo.utils import fullname
from diazo.utils import namespaces
from diazo.utils import pkg_xsl
from lxml import etree
from urllib.parse import urljoin
from urllib.request import urlopen

import logging
import re


logger = logging.getLogger("diazo")
usage = __doc__

IMPORT_STYLESHEET = re.compile(
    r"""(?P<before>@import[ \t]+(?P<paren>url\([ \t]?)?(?P<quote>['"]?))"""
    r"""(?P<url>\S+)(?P<after>(?P=quote)(?(paren)\)))""",
    re.IGNORECASE,
)
CONDITIONAL_SRC = re.compile(
    r"""(?P<before><[^>]*?(src|href)=(?P<quote>['"]?))"""
    r"""(?P<url>[^ \t\n\r\f\v>]+)(?P<after>(?P=quote)[^>]*?>)""",
    re.IGNORECASE,
)
SRCSET = re.compile(r"(?P<descriptors>^\s*|\s*,\s*)(?P<url>[^\s]*)")


update_transform = pkg_xsl("update-namespace.xsl")
normalize_rules = pkg_xsl("normalize-rules.xsl")
include = pkg_xsl("include.xsl")
apply_conditions = pkg_xsl("apply-conditions.xsl")
merge_conditions = pkg_xsl("merge-conditions.xsl")
annotate_themes = pkg_xsl("annotate-themes.xsl")
annotate_rules = pkg_xsl("annotate-rules.xsl")
apply_rules = pkg_xsl("apply-rules.xsl")
fixup_themes = pkg_xsl("fixup-themes.xsl")


def anchor_safe_urljoin(base, url):
    """Join the base with the url only when the url doesn't start with '#'"""
    if url.startswith("#"):
        return url
    else:
        return urljoin(base, url)


def add_identifiers(rules_doc):
    """Add identifiers to the rules for debugging"""
    for index, elem in enumerate(
        rules_doc.xpath(
            "//diazo:rules | //diazo:rules/diazo:*"
            " | //old1:rules | //old1:rules/old1:*"
            " | //old2:rules | //old2:rules/old1:*",
            namespaces=namespaces,
        ),
    ):
        elem.set(
            fullname(
                namespaces["xml"],
                "id",
            ),
            f"r{index}",
        )
    return rules_doc


def update_namespace(rules_doc):
    """Convert old namespace to new namespace in place"""
    update = False
    for ns in (namespaces["old1"], namespaces["old2"]):
        if rules_doc.xpath(f"//*[namespace-uri()='{ns:s}']"):
            logger.warning(
                "The %s namespace is deprecated, use %s instead.",
                ns,
                namespaces["diazo"],
            )
            update = True
    for ns in (namespaces["oldcss1"], namespaces["oldcss2"]):
        if rules_doc.xpath(f"//@*[namespace-uri()='{ns:s}']"):
            logger.warning(
                "The %s namespace is deprecated, use %s instead.",
                ns,
                namespaces["css"],
            )
            update = True
    if update:
        new_doc = update_transform(rules_doc)
        # Place the nodes into the old tree to preserve any custom resolvers
        new = new_doc.getroot()
        root = rules_doc.getroot()
        root.clear()
        root.tag = new.tag
        root.nsmap.update(new.nsmap.items())
        root.attrib.update(new.attrib.items())
        root.text = new.text
        root[:] = new[:]
        root.tail = new.tail
    return rules_doc


def escape_curly_brackets(theme_doc):
    for node in theme_doc.iter():
        for attr in node.attrib:
            if "{" in node.attrib[attr]:
                node.attrib[attr] = node.attrib[attr].replace("{", "{{")
            if "}" in node.attrib[attr]:
                node.attrib[attr] = node.attrib[attr].replace("}", "}}")


def expand_theme(element, theme_doc, absolute_prefix):
    prefix = urljoin(absolute_prefix, element.get("prefix", ""))
    apply_absolute_prefix(theme_doc, prefix)
    escape_curly_brackets(theme_doc)
    theme_root = theme_doc.getroot()
    preceding = list(theme_root.itersiblings(preceding=True))
    preceding.reverse()
    following = list(theme_root.itersiblings(preceding=False))
    element.extend(preceding)
    element.append(theme_root)
    element.extend(following)


def expand_themes(
    rules_doc,
    parser=None,
    absolute_prefix=None,
    read_network=False,
):
    """Expand <theme href='...'/> nodes with the theme html."""
    if absolute_prefix is None:
        absolute_prefix = ""
    base = rules_doc.docinfo.URL
    if parser is None:
        parser = etree.HTMLParser()
    for element in rules_doc.xpath(
        "//diazo:theme[@href]",
        namespaces=namespaces,
    ):
        url = urljoin(base, element.get("href"))
        if not read_network and url.startswith(
            ("ftp://", "ftps://", "http://", "https://")
        ):
            raise ValueError(
                "Supplied theme '{url}', but network access denied.".format(
                    url=url,
                ),
            )
        elif read_network and url.startswith(
            ("ftp://", "ftps://", "http://", "https://")
        ):
            theme = urlopen(url)
        else:
            theme = url
        theme_doc = etree.parse(theme, parser=parser, base_url=url)
        expand_theme(element, theme_doc, absolute_prefix)
    return rules_doc


def apply_absolute_prefix(theme_doc, absolute_prefix):
    if not absolute_prefix:
        return
    if not absolute_prefix.endswith("/"):
        absolute_prefix = absolute_prefix + "/"
    for node in theme_doc.xpath("//*[@src]"):
        url = urljoin(absolute_prefix, node.get("src"))
        node.set("src", url)
    for xlink_attr in theme_doc.xpath('//@*[local-name()="xlink:href"]'):
        node = xlink_attr.getparent()
        url = anchor_safe_urljoin(absolute_prefix, node.get("xlink:href"))
        node.set("xlink:href", url)
    for node in theme_doc.xpath("//*[@srcset]"):
        srcset = node.get("srcset")
        srcset = SRCSET.sub(
            lambda match: match.group("descriptors")
            + urljoin(
                absolute_prefix,
                match.group("url"),
            ),
            srcset,
        )
        node.set("srcset", srcset)
    for node in theme_doc.xpath("//*[@href]"):
        url = anchor_safe_urljoin(absolute_prefix, node.get("href"))
        node.set("href", url)
    for node in theme_doc.xpath("//style"):
        if node.text is None:
            continue
        node.text = IMPORT_STYLESHEET.sub(
            lambda match: match.group("before")
            + urljoin(
                absolute_prefix,
                match.group("url"),
            )
            + match.group("after"),
            node.text,
        )
    for node in theme_doc.xpath('//comment()[starts-with(., "[if")]'):
        node.text = IMPORT_STYLESHEET.sub(
            lambda match: match.group("before")
            + urljoin(
                absolute_prefix,
                match.group("url"),
            )
            + match.group("after"),
            node.text,
        )
        node.text = CONDITIONAL_SRC.sub(
            lambda match: match.group("before")
            + urljoin(
                absolute_prefix,
                match.group("url"),
            )
            + match.group("after"),
            node.text,
        )


def add_extra(rules_doc, extra):
    root = rules_doc.getroot()
    extra_elements = extra.xpath(
        "/xsl:stylesheet/xsl:*",
        namespaces=namespaces,
    )
    root.extend(extra_elements)
    return rules_doc


def add_theme(
    rules_doc,
    theme,
    parser=None,
    absolute_prefix=None,
    read_network=False,
):
    if (
        not read_network
        and isinstance(theme, str)
        and theme[:6] in ("ftp://", "http:/", "https:")
    ):
        raise ValueError(
            "Supplied theme '{theme}', but network access denied.".format(
                theme=theme,
            ),
        )
    if absolute_prefix is None:
        absolute_prefix = ""
    if parser is None:
        parser = etree.HTMLParser()
    root = rules_doc.getroot()
    element = root.makeelement(fullname(namespaces["diazo"], "theme"))
    root.append(element)
    theme_doc = etree.parse(theme, parser=parser)
    expand_theme(element, theme_doc, absolute_prefix)
    return rules_doc


def fixup_theme_comment_selectors(rules):
    """Comments must be converted to <xsl:comment> to be output, doing it early
    allows them to get an xml:id so they can be matched in the theme. The theme
    selector needs rewriting to replace comment() with xsl:comment
    """
    for element in rules.xpath("//@theme[contains(., 'comment()')]/.."):
        element.attrib["theme"] = element.attrib["theme"].replace(
            "comment()",
            "xsl:comment",
        )
    return rules


def process_rules(
    rules,
    theme=None,
    extra=None,
    trace=None,
    css=True,
    xinclude=True,
    absolute_prefix=None,
    includemode=None,
    update=True,
    parser=None,
    rules_parser=None,
    read_network=False,
    stop=None,
):
    if trace:
        trace = "1"
    else:
        trace = "0"
    if rules_parser is None:
        rules_parser = etree.XMLParser(recover=False)
    rules_doc = etree.parse(rules, parser=rules_parser)
    if stop == 0:
        return rules_doc
    if parser is None:
        parser = etree.HTMLParser()
    if xinclude:
        # XXX: read_network limitation not yet supported  # NOQA: T000
        #   for xinclude
        rules_doc.xinclude()
    if stop == 1:
        return rules_doc
    rules_doc = add_identifiers(rules_doc)
    if stop == 2 or stop == "add_identifiers":
        return rules_doc
    if update:
        rules_doc = update_namespace(rules_doc)
    if stop == 3:
        return rules_doc
    if css:
        rules_doc = convert_css_selectors(rules_doc)
    if stop == 4:
        return rules_doc
    rules_doc = fixup_theme_comment_selectors(rules_doc)
    if stop == 5:
        return rules_doc
    rules_doc = expand_themes(rules_doc, parser, absolute_prefix, read_network)
    if theme is not None:
        rules_doc = add_theme(
            rules_doc,
            theme,
            parser,
            absolute_prefix,
            read_network,
        )
    if stop == 6:
        return rules_doc
    if includemode is None:
        includemode = "document"
    includemode = f"'{includemode:s}'"
    rules_doc = normalize_rules(rules_doc, includemode=includemode)
    if stop == 7:
        return rules_doc
    rules_doc = apply_conditions(rules_doc)
    if stop == 8:
        return rules_doc
    rules_doc = merge_conditions(rules_doc)
    if stop == 9:
        return rules_doc
    rules_doc = fixup_themes(rules_doc)
    if stop == 10:
        return rules_doc
    rules_doc = annotate_themes(rules_doc)
    if stop == 11:
        return rules_doc
    rules_doc = include(rules_doc)
    if stop == 12:
        return rules_doc
    rules_doc = annotate_rules(rules_doc)
    if stop == 13:
        return rules_doc
    rules_doc = apply_rules(rules_doc, trace=trace)
    return rules_doc


def main():
    """Called from console script"""
    parser = _createOptionParser(usage=usage)
    parser.add_option(
        "-s",
        "--stop",
        metavar="n",
        type="int",
        help="Stop preprocessing at stage n",
        dest="stop",
        default=None,
    )
    (options, args) = parser.parse_args()

    if options.rules is None:
        if len(args) == 2 and options.theme is None:
            options.rules, options.theme = args
        elif len(args) == 1:
            (options.rules,) = args
        else:
            parser.error("Wrong number of arguments.")
    elif args:
        parser.error("Wrong number of arguments.")

    if options.trace:
        logger.setLevel(logging.DEBUG)

    rules_doc = process_rules(
        options.rules,
        theme=options.theme,
        extra=options.extra,
        trace=options.trace,
        absolute_prefix=options.absolute_prefix,
        includemode=options.includemode,
        read_network=options.read_network,
        stop=options.stop,
    )
    root = rules_doc.getroot()
    if not root.tail:
        root.tail = "\n"
    rules_doc.write(options.output, pretty_print=options.pretty_print)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""\
Usage: %prog [options] [-r] RULES [-t] THEME

  THEME is an html file.
  RULES is a file defining a set of diazo rules in css syntax, e.g:
    <rules xmlns="http://namespaces.plone.org/diazo">
        <copy content="//div[@id='content-wrapper']"
              theme="//div[@id='page-content']"/>
    </rules>\
"""

from diazo.rules import process_rules
from diazo.utils import _createOptionParser
from diazo.utils import CustomResolver
from diazo.utils import pkg_xsl
from diazo.utils import quote_param
from diazo.utils import split_params
from importlib.resources import files
from lxml import etree

import logging


logger = logging.getLogger("diazo")
usage = __doc__


def set_parser(stylesheet, parser, compiler_parser=None):
    file_obj = files("diazo") / "dummy.html"
    with open(file_obj) as file_handler:
        dummy_doc = etree.parse(
            file_handler,
            parser=parser,
        )
    name = "file:///__diazo__"
    resolver = CustomResolver({name: stylesheet})
    if compiler_parser is None:
        compiler_parser = etree.XMLParser()
    compiler_parser.resolvers.add(resolver)
    identity = pkg_xsl("identity.xsl", compiler_parser)
    output_doc = identity(dummy_doc, docurl=f"'{name}'")
    compiler_parser.resolvers.remove(resolver)
    return output_doc


def build_xsl_params_document(xsl_params):
    if xsl_params is None:
        xsl_params = {}
    if "path" not in xsl_params:
        xsl_params["path"] = ""
    known_params = etree.XML(
        '<xsl:stylesheet version="1.0" '
        'xmlns:xsl="http://www.w3.org/1999/XSL/Transform" />',
    )
    for param_name, param_value in xsl_params.items():
        param_element = etree.SubElement(
            known_params,
            "{http://www.w3.org/1999/XSL/Transform}param",
        )
        param_element.attrib["name"] = param_name
        if isinstance(param_value, str):
            param_element.text = param_value
        else:
            param_element.attrib["select"] = str(quote_param(param_value))
        param_element.tail = "\n"

    return known_params


def compile_theme(
    rules,
    theme=None,
    extra=None,
    css=True,
    xinclude=True,
    absolute_prefix=None,
    update=True,
    trace=False,
    includemode=None,
    parser=None,
    compiler_parser=None,
    rules_parser=None,
    access_control=None,
    read_network=False,
    indent=None,
    xsl_params=None,
    runtrace=False,
):
    """Invoke the diazo compiler.

    * ``rules`` is the rules file
    * ``theme`` is the theme file
    * ``extra`` is an optional XSLT file with Diazo extensions (deprecated, use
      inline xsl in the rules instead)
    * ``css``   can be set to False to disable CSS syntax support (providing a
      moderate speed gain)
    * ``xinclude`` can be set to False to disable XInclude support during the
      compile phase (providing a moderate speed gain)
    * ``absolute_prefix`` can be set to a string that will be prefixed to any
      *relative* URL referenced in an image, link or stylesheet in the theme
      HTML file before the theme is passed to the compiler. This allows a
      theme to be written so that it can be opened and views standalone on the
      filesystem, even if at runtime its static resources are going to be
      served from some other location. For example, an
      ``<img src="images/foo.jpg" />`` can be turned into
      ``<img src="/static/images/foo.jpg" />`` with an ``absolute_prefix`` of
      "/static".
    * ``update`` can be set to False to disable the automatic update support
      for the old Deliverance 0.2 namespace (for a moderate speed gain)
    * ``trace`` can be set to True to enable compiler trace information
    * ``runtrace`` can be set to True to add tracing into the XSL output
    * ``includemode`` can be set to 'document', 'esi' or 'ssi' to change the
      way in which includes are processed
    * ``parser`` can be set to an lxml parser instance; the default is an
      HTMLParser
    * ``compiler_parser``` can be set to an lxml parser instance; the default
      is a XMLParser
    * ``rules_parser`` can be set to an lxml parser instance; the default is a
      XMLParser.
    * ``xsl_params`` can be set to a dictionary of parameters that will be
      known to the compiled theme transform. The keys should be the parameter
      names. Values are default values.
    """
    if access_control is not None:
        read_network = access_control.options["read_network"]
    rules_doc = process_rules(
        rules=rules,
        theme=theme,
        extra=extra,
        css=css,
        xinclude=xinclude,
        absolute_prefix=absolute_prefix,
        update=update,
        trace=trace,
        includemode=includemode,
        parser=parser,
        rules_parser=rules_parser,
        read_network=read_network,
    )

    # Build a document with all the <xsl:param /> values to set the defaults
    # for every value passed in as xsl_params
    known_params = build_xsl_params_document(xsl_params)

    # Create a pseudo resolver for this
    known_params_url = "file:///__diazo_known_params__"
    emit_stylesheet_resolver = CustomResolver(
        {
            known_params_url: etree.tostring(known_params),
        }
    )
    emit_stylesheet_parser = etree.XMLParser()
    emit_stylesheet_parser.resolvers.add(emit_stylesheet_resolver)

    # Set up parameters
    params = {}
    if indent is not None:
        params["indent"] = indent and "'yes'" or "'no'"
    params["known_params_url"] = quote_param(known_params_url)
    params["runtrace"] = "1" if runtrace else "0"

    # Run the final stage compiler
    emit_stylesheet = pkg_xsl(
        "emit-stylesheet.xsl",
        parser=emit_stylesheet_parser,
    )
    compiled_doc = emit_stylesheet(rules_doc, **params)
    compiled_doc = set_parser(
        etree.tostring(compiled_doc),
        parser,
        compiler_parser,
    )

    return compiled_doc


def main():
    """Called from console script"""
    parser = _createOptionParser(usage=usage)
    (options, args) = parser.parse_args()

    if options.rules is None:
        if len(args) == 2 and options.theme is None:
            options.rules, options.theme = args
        elif len(args) == 1:
            (options.rules,) = args
        else:
            parser.error("Wrong number of arguments.")
    elif args:
        parser.error("Wrong number of arguments.")

    if options.trace:
        logger.setLevel(logging.DEBUG)

    xsl_params = None
    if options.xsl_params:
        xsl_params = split_params(options.xsl_params)

    output_xslt = compile_theme(
        rules=options.rules,
        theme=options.theme,
        extra=options.extra,
        trace=options.trace,
        absolute_prefix=options.absolute_prefix,
        includemode=options.includemode,
        read_network=options.read_network,
        xsl_params=xsl_params,
    )
    root = output_xslt.getroot()
    if not root.tail:
        root.tail = "\n"
    output_xslt.write(
        options.output,
        encoding="utf-8",
        pretty_print=options.pretty_print,
    )


if __name__ == "__main__":
    main()



from io import BytesIO
from lxml import etree

import diazo.compiler
import diazo.run
import diazo.runtrace
import os.path
import sys
import unittest


if __name__ == "__main__":
    __file__ = sys.argv[0]


def _testfile(filename):
    return os.path.join(
        os.path.abspath(os.path.dirname(__file__)),
        "test_wsgi_files",
        filename,
    )


class TestDebug(unittest.TestCase):
    rules_str = b"""\
<rules xmlns="http://namespaces.plone.org/diazo"
       xmlns:css="http://namespaces.plone.org/diazo/css"
       xmlns:xsl="http://www.w3.org/1999/XSL/Transform">
  <theme css:if-content="body.external">
    <html><body>
      <h1>External Theme</h1>
      <div class="cow">I am daisy the cow</div>
      <div class="pig">I am daisy the pig</div>
    </body></html>
  </theme>
  <rules if-content="/html/body[@id = 'theme-on']"
         useless="I need to be put before if-content processing">
      <replace css:content="div.bovine"
               css:theme="div.cow"
               css:if-content="body.female" />
      <replace css:content="div.bovine"
               css:theme="div.bull"
               css:if-content="body.male" />
      <replace css:content="div.pig" css:theme="div.pig" />
      <replace css:content="div.antelope" css:theme="div.antelope" />
      <replace content='//*[@id="some_other_node_but_weird_quoting"]'
               css:theme-children='#alpha' />
      <replace css:content="div.iguana" css:theme="div.bull"
               css:if-not-content="body.male" />
  </rules>
</rules>
    """
    theme_str = b"""\
<html><head>
  <meta http-equiv="content-type"
        content="text/html; charset=utf-8; i-am-not-a-diazo-rule" />
</head><body>
  <h1>Provided Theme</h1>
  <div class="cow">I am a template cow</div>
  <div class="bull">I am a template bull</div>
  <div class="pig">I am daisy the pig</div>
</body></html>
    """

    def compile(self):
        # Compile default rule and themes
        ct = diazo.compiler.compile_theme(
            rules=BytesIO(self.rules_str),
            theme=BytesIO(self.theme_str),
            indent=True,
            runtrace=True,
        )
        return etree.XSLT(ct)

    def test_internal(self):
        processor = self.compile()
        processor(
            etree.fromstring(
                """\
<html><body id="theme-on" class="male">
  <h1>Content</h1>
  <div class="bovine" id="#cow-daisy">I am frank the bull</div>
  <div class="pig" id="#pig-george">I am daisy the pig</div>
</body></html>
""",
            ),
        )
        runtrace_doc = diazo.runtrace.generate_runtrace(
            rules=BytesIO(self.rules_str),
            error_log=processor.error_log,
        )
        self.assertXPath(
            runtrace_doc,
            "/d:rules/d:theme/@runtrace-if-content",
            "false",
        )
        self.assertXPath(
            runtrace_doc,
            "/d:rules/d:rules/@runtrace-if-content",
            "true",
        )
        # <replace css:content="div.bovine"
        #          css:theme="div.cow"
        #          css:if-content="body.female" />
        self.assertXPath(
            runtrace_doc,
            "/d:rules/d:rules/d:replace[1]/@runtrace-if-content",
            "false",
        )
        self.assertXPath(
            runtrace_doc,
            "/d:rules/d:rules/d:replace[1]/@runtrace-content",
            "1",
        )
        self.assertXPath(
            runtrace_doc,
            "/d:rules/d:rules/d:replace[1]/@runtrace-theme",
            "1",
        )
        self.assertXPath(
            runtrace_doc,
            "/d:rules/d:rules/d:replace[1]/@runtrace-merged-condition",
            "false",
        )
        # <replace css:content="div.bovine"
        #          css:theme="div.bull"
        #          css:if-content="body.male" />
        self.assertXPath(
            runtrace_doc,
            "/d:rules/d:rules/d:replace[2]/@runtrace-if-content",
            "true",
        )
        self.assertXPath(
            runtrace_doc,
            "/d:rules/d:rules/d:replace[2]/@runtrace-content",
            "1",
        )
        self.assertXPath(
            runtrace_doc,
            "/d:rules/d:rules/d:replace[2]/@runtrace-theme",
            "1",
        )
        self.assertXPath(
            runtrace_doc,
            "/d:rules/d:rules/d:replace[2]/@runtrace-merged-condition",
            "true",
        )
        # <replace css:content="div.pig" css:theme="div.pig" />
        self.assertXPath(
            runtrace_doc,
            "/d:rules/d:rules/d:replace[3]/@runtrace-content",
            "1",
        )
        self.assertXPath(
            runtrace_doc,
            "/d:rules/d:rules/d:replace[3]/@runtrace-merged-condition",
            "true",
        )
        # <replace css:content="div.antelope" css:theme="div.antelope" />
        self.assertXPath(
            runtrace_doc,
            "/d:rules/d:rules/d:replace[4]/@runtrace-content",
            "0",
        )
        self.assertXPath(
            runtrace_doc,
            "/d:rules/d:rules/d:replace[4]/@runtrace-merged-condition",
            "true",
        )
        # <replace css:content="div.iguana" css:theme="div.bull"
        #          css:if-not-content="body.male" />
        self.assertXPath(
            runtrace_doc,
            "/d:rules/d:rules/d:replace[6]/@runtrace-if-not-content",
            "false",
        )
        self.assertXPath(
            runtrace_doc,
            "/d:rules/d:rules/d:replace[6]/@runtrace-merged-condition",
            "false",
        )

    def test_external(self):
        processor = self.compile()
        processor(
            etree.fromstring(
                """\
<html><body id="theme-on" class="female external">
  <h1>Content</h1>
  <div class="bovine" id="#cow-daisy">I am daisy the cow</div>
  <div class="pig" id="#pig-george">I am daisy the pig</div>
</body></html>
""",
            ),
        )
        runtrace_doc = diazo.runtrace.generate_runtrace(
            rules=BytesIO(self.rules_str),
            error_log=processor.error_log,
        )
        self.assertXPath(
            runtrace_doc,
            "/d:rules/d:theme/@runtrace-if-content",
            "true",
        )
        self.assertXPath(
            runtrace_doc,
            "/d:rules/d:rules/@runtrace-if-content",
            "true",
        )
        # <replace css:content="div.bovine"
        #          css:theme="div.cow"
        #          css:if-content="body.female" />
        self.assertXPath(
            runtrace_doc,
            "/d:rules/d:rules/d:replace[1]/@runtrace-if-content",
            "true",
        )
        self.assertXPath(
            runtrace_doc,
            "/d:rules/d:rules/d:replace[1]/@runtrace-content",
            "1",
        )
        self.assertXPath(
            runtrace_doc,
            "/d:rules/d:rules/d:replace[1]/@runtrace-theme",
            "1",
        )
        self.assertXPath(
            runtrace_doc,
            "/d:rules/d:rules/d:replace[1]/@runtrace-merged-condition",
            "true",
        )
        # <replace css:content="div.bovine"
        #          css:theme="div.bull"
        #          css:if-content="body.male" />
        # The external theme only has the cow slot
        self.assertXPath(
            runtrace_doc,
            "/d:rules/d:rules/d:replace[2]/@runtrace-if-content",
            "false",
        )
        self.assertXPath(
            runtrace_doc,
            "/d:rules/d:rules/d:replace[2]/@runtrace-content",
            "1",
        )
        self.assertXPath(
            runtrace_doc,
            "/d:rules/d:rules/d:replace[2]/@runtrace-theme",
            "0",
        )
        self.assertXPath(
            runtrace_doc,
            "/d:rules/d:rules/d:replace[2]/@runtrace-merged-condition",
            "false",
        )
        # <replace css:content="div.pig" css:theme="div.pig" />
        self.assertXPath(
            runtrace_doc,
            "/d:rules/d:rules/d:replace[3]/@runtrace-content",
            "1",
        )
        self.assertXPath(
            runtrace_doc,
            "/d:rules/d:rules/d:replace[3]/@runtrace-merged-condition",
            "true",
        )
        # <replace css:content="div.antelope" css:theme="div.antelope" />
        self.assertXPath(
            runtrace_doc,
            "/d:rules/d:rules/d:replace[4]/@runtrace-content",
            "0",
        )
        self.assertXPath(
            runtrace_doc,
            "/d:rules/d:rules/d:replace[4]/@runtrace-merged-condition",
            "true",
        )
        # <replace css:content="div.iguana" css:theme="div.bull"
        #          css:if-not-content="body.male" />
        self.assertXPath(
            runtrace_doc,
            "/d:rules/d:rules/d:replace[6]/@runtrace-if-not-content",
            "true",
        )
        self.assertXPath(
            runtrace_doc,
            "/d:rules/d:rules/d:replace[6]/@runtrace-merged-condition",
            "true",
        )

    def test_htmlformat(self):
        html_string = etree.tostring(
            diazo.runtrace.runtrace_to_html(
                etree.fromstring(
                    """\
<rules xmlns="http://namespaces.plone.org/diazo"
       xmlns:css="http://namespaces.plone.org/diazo/css"
       xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
       css:if-content="#visual-portal-wrapper"
       xml:id="r0" runtrace-if-content="true">

    <theme href="index.html" xml:id="r1"/>
    <notheme if-path="presentation_view" xml:id="r2"/>
    <notheme if-path="source_editor.htm" xml:id="r3"/>
    <rules xml:id="r4">
        <!-- Rules, lots of rules -->
        <copy xml:id="r5" content="//a" theme="//a" runtrace-content="0"
              runtrace-theme="0" />
        <copy xml:id="r6" content="//a" theme="//b" runtrace-content="1"
              runtrace-theme="0" />
        <copy xml:id="r7" content="//b" theme="//b" runtrace-content="1"
              runtrace-theme="1" />
        <copy xml:id="r8" content="//b" theme="//c" runtrace-content="1"
              runtrace-theme="2" />
    </rules>
</rules>
""",
                ),
            ),
        )
        # First rule has an if-content condition
        self.assertIn(
            b"""<pre class="runtrace"><span class="node match" """
            b"""title="Matches: if-content:true ">&lt;rules""",
            html_string,
        )
        # HTML comments are included and escaped
        self.assertIn(
            b"""&lt;!-- Rules, lots of rules --&gt;""",
            html_string,
        )
        # Rules tag has children
        self.assertIn(
            b"""<span class="node unrelated">&lt;rules """
            b"""<span class="attr">xml:id="r4"</span>&gt;</span>""",
            html_string,
        )
        # Theme tag has no conditions, is a singleton
        self.assertIn(
            b"""<span class="node unrelated">&lt;theme <span """
            b"""class="attr">href="index.html"</span> <span """
            b"""class="attr">xml:id="r1"</span>/&gt;</span>""",
            html_string,
        )
        # Whitespace is preserved
        self.assertIn(
            b"""xml:id=\"r4\"</span>&gt;</span>\n        <span """
            b"""class="comment">&lt;!-- Rules, lots of rules """
            b"""--&gt;</span>""",
            html_string,
        )
        # Neither theme or content matched
        self.assertIn(
            b"""<span class="node no-match" title="Matches: """
            b"""content:0 theme:0 ">&lt;copy <span class="attr">"""
            b"""xml:id="r5"</span>""",
            html_string,
        )
        # Just content matched, still not good enough
        self.assertIn(
            b"""<span class="node no-match" title="Matches: """
            b"""content:1 theme:0 ">&lt;copy <span class="attr">"""
            b"""xml:id="r6"</span>""",
            html_string,
        )
        # Full match
        self.assertIn(
            b"""<span class="node match" title="Matches: """
            b"""content:1 theme:1 ">&lt;copy <span class="attr">"""
            b"""xml:id="r7"</span>""",
            html_string,
        )
        # More than one match still fine
        self.assertIn(
            b"""<span class="node match" title="Matches: """
            b"""content:1 theme:2 ">&lt;copy <span class="attr">"""
            b"""xml:id="r8"</span>""",
            html_string,
        )

    def assertXPath(self, doc, xpath, expected):
        self.assertEqual(
            doc.xpath(
                xpath,
                namespaces=(dict(d="http://namespaces.plone.org/diazo")),
            )[0],
            expected,
        )


def test_suite():
    return unittest.defaultTestLoader.loadTestsFromName(__name__)



from diazo.compiler import compile_theme
from diazo.utils import pkg_parse
from diazo.utils import quote_param
from importlib.resources import files
from lxml import etree
from repoze.xmliter.serializer import XMLSerializer
from repoze.xmliter.utils import getHTMLSerializer
from urllib.parse import unquote_plus
from webob import Request

import os.path
import re


DIAZO_OFF_HEADER = "X-Diazo-Off"


def asbool(value):
    if isinstance(value, str):
        value = value.strip().lower()
        if value in (
            "true",
            "yes",
            "on",
            "y",
            "t",
            "1",
        ):
            return True
        elif value in (
            "false",
            "no",
            "off",
            "n",
            "f",
            "0",
        ):
            return False
        else:
            raise ValueError(
                "String is not true/false: {value:r}".format(
                    value=value,
                ),
            )
    else:
        return bool(value)


class FilesystemResolver(etree.Resolver):
    """Resolver for filesystem paths"""

    def resolve(self, system_url, public_id, context):
        if "://" not in system_url and os.path.exists(system_url):
            return self.resolve_filename(system_url, context)
        else:
            return None


class NetworkResolver(etree.Resolver):
    """Resolver for network urls"""

    def resolve(self, system_url, public_id, context):
        if "://" in system_url and system_url != "file:///__diazo__":
            return self.resolve_filename(system_url, context)
        else:
            return None


class PythonResolver(etree.Resolver):
    """Resolver for python:// urls"""

    def resolve(self, system_url, public_id, context):
        if not system_url.lower().startswith("python://"):
            return None

        spec = system_url[9:]
        package, resource_name = spec.split("/", 1)
        filename = str(files(package) / resource_name)

        return self.resolve_filename(filename, context)


class WSGIResolver(etree.Resolver):
    """Resolver that performs a WSGI subrequest"""

    def __init__(self, app):
        self.app = app

    def resolve(self, system_url, public_id, context):
        # Ignore URLs with a scheme
        if "://" in system_url:
            return None

        # Ignore the special 'diazo:' resolvers
        if system_url.startswith("diazo:"):
            return None

        subrequest = Request.blank(system_url)
        response = subrequest.get_response(self.app)

        status_code, reason = response.status.split(None, 1)
        if not status_code == "200":
            return None

        if response.charset is None:
            response.charset = "UTF-8"  # Maybe this should be latin1?

        result = response.text

        if response.content_type in (
            "text/javascript",
            "application/x-javascript",
        ):
            result = "".join(
                [
                    '<html><body><script type="text/javascript">',
                    result,
                    "</script></body></html>",
                ]
            )
        elif response.content_type == "text/css":
            result = "".join(
                [
                    '<html><body><style type="text/css">',
                    result,
                    "</style></body></html>",
                ]
            )

        return self.resolve_string(result, context)


class XSLTMiddleware:
    """Apply XSLT in middleware"""

    def __init__(
        self,
        app,
        global_conf,
        filename=None,
        tree=None,
        read_network=False,
        read_file=True,
        update_content_length=False,
        ignored_extensions=(
            "js",
            "css",
            "gif",
            "jpg",
            "jpeg",
            "pdf",
            "ps",
            "doc",
            "png",
            "ico",
            "mov",
            "mpg",
            "mpeg",
            "mp3",
            "m4a",
            "txt",
            "rtf",
            "swf",
            "wav",
            "zip",
            "wmv",
            "ppt",
            "gz",
            "tgz",
            "jar",
            "xls",
            "bmp",
            "tif",
            "tga",
            "hqx",
            "avi",
        ),
        environ_param_map=None,
        unquoted_params=None,
        doctype=None,
        content_type=None,
        charset=None,
        remove_conditional_headers=False,
        **params,  # NOQA: C816
    ):
        """Initialise, giving a filename or parsed XSLT tree.

        The parameters are:

        * ``filename``, a filename from which to read the XSLT file
        * ``tree``, a pre-parsed lxml tree representing the XSLT file

        ``filename`` and ``tree`` are mutually exclusive.

        * ``read_network``, should be set to True to allow resolving resources
          from the network.
        * ``read_file``, should be set to False to disallow resolving resources
          from the filesystem.
        * ``update_content_length``, can be set to True to update the
          Content-Length header when applying the transformation. When set to
          False (the default), the header is removed and it is left to the WSGI
          server recalculate or send a chunked response.
        * ``ignored_extensions`` can be set to a list of filename extensions
          for which the transformation should never be applied
        * ``environ_param_map`` can be set to a dict of environ keys to
          parameter names. The corresponding values will then be sent to the
          transformation as parameters.
        * ``unquoted_params``, can be set to a list of parameter names which
          will not be quoted.
        * ``doctype``, can be set to a string which will replace that set in
          the XSLT, for example, "<!DOCTYPE html>".
        * ``content_type``, can be set to a string which will be set in the
          Content-Type header. By default it is inferred from the stylesheet.
        * ``charset``, can be set to a string which will be set in the
          Content-Type header. By default it is inferred from the stylesheet.
        * ``remove_conditional_headers``, should be set to True if the
        transformed output includes other files.

        Additional keyword arguments will be passed to the transformation as
        parameters.
        """

        self.app = app
        self.global_conf = global_conf

        if filename is not None:
            xslt_file = open(filename)
            source = xslt_file.read()
            tree = etree.fromstring(source)
            xslt_file.close()

        if content_type is None:
            mediatype = tree.xpath(
                "/xsl:stylesheet/xsl:output/@media-type",
                namespaces=dict(xsl="http://www.w3.org/1999/XSL/Transform"),
            )
            if mediatype:
                content_type = mediatype[-1]
            else:
                method = tree.xpath(
                    "/xsl:stylesheet/xsl:output/@method",
                    namespaces=dict(
                        xsl="http://www.w3.org/1999/XSL/Transform",
                    ),
                )
                if method:
                    method = method[-1]
                    if method.lower() == "html":
                        content_type = "text/html"
                    elif method.lower() == "text":
                        content_type = "text/plain"
                    elif method.lower() == "xml":
                        content_type = "text/xml"
        self.content_type = content_type

        if charset is None:
            encoding = tree.xpath(
                "/xsl:stylesheet/xsl:output/@encoding",
                namespaces=dict(xsl="http://www.w3.org/1999/XSL/Transform"),
            )
            if encoding:
                charset = encoding[-1]
            else:
                charset = "UTF-8"
        self.charset = charset

        self.read_network = asbool(read_network)
        self.read_file = asbool(read_file)
        self.access_control = etree.XSLTAccessControl(
            read_file=self.read_file,
            write_file=False,
            create_dir=False,
            read_network=self.read_network,
            write_network=False,
        )
        self.transform = etree.XSLT(tree, access_control=self.access_control)
        self.update_content_length = asbool(update_content_length)
        self.ignored_extensions = frozenset(ignored_extensions)

        self.ignored_pattern = re.compile(
            r"^.*\.({ext:s})$".format(ext="|".join(ignored_extensions)),
        )

        self.environ_param_map = environ_param_map or {}
        if isinstance(unquoted_params, str):
            unquoted_params = unquoted_params.split()
        self.unquoted_params = unquoted_params and frozenset(unquoted_params) or ()
        self.params = params
        self.doctype = doctype
        self.remove_conditional_headers = asbool(remove_conditional_headers)

    def __call__(self, environ, start_response):
        request = Request(environ)
        if self.should_ignore(request):
            return self.app(environ, start_response)

        if self.remove_conditional_headers:
            request.remove_conditional_headers()
        else:
            # Always remove Range and Accept-Encoding headers
            request.remove_conditional_headers(
                remove_encoding=True,
                remove_range=False,
                remove_match=False,
                remove_modified=True,
            )

        response = request.get_response(self.app)
        if not self.should_transform(response):
            return response(environ, start_response)
        try:
            input_encoding = response.charset

            # Note, the Content-Length header will not be set
            if request.method == "HEAD":
                self.reset_headers(response)
                return response(environ, start_response)

            # Prepare the serializer
            try:
                serializer = getHTMLSerializer(
                    response.app_iter,
                    encoding=input_encoding,
                )
            except etree.XMLSyntaxError:
                # Abort transform on syntax error for empty response
                # Headers should be left intact
                return response(environ, start_response)
        finally:
            if getattr(response.app_iter, "close", None):
                response.app_iter.close()

        self.reset_headers(response)

        # Set up parameters

        params = {}
        for key, value in self.environ_param_map.items():
            if key in environ:
                if value in self.unquoted_params:
                    params[value] = environ[key]
                else:
                    params[value] = quote_param(environ[key])
        for key, value in self.params.items():
            if key in self.unquoted_params:
                params[key] = value
            else:
                params[key] = quote_param(value)

        # Apply the transformation
        tree = self.transform(serializer.tree, **params)

        # Set content type (normally inferred from stylesheet)
        # Unfortunately lxml does not expose docinfo.mediaType
        if self.content_type is None:
            if tree.getroot().tag == "html":
                response.content_type = "text/html"
            else:
                response.content_type = "text/xml"
        response.charset = tree.docinfo.encoding or self.charset

        # Return a repoze.xmliter XMLSerializer, which helps avoid re-parsing
        # the content tree in later middleware stages.
        response.app_iter = XMLSerializer(tree, doctype=self.doctype)

        # Calculate the content length - we still return the parsed tree
        # so that other middleware could avoid having to re-parse, even if
        # we take a hit on serialising here
        if self.update_content_length:
            response.content_length = len(bytes(response.app_iter))

        return response(environ, start_response)

    def should_ignore(self, request):
        """Determine if we should ignore the request"""

        if asbool(request.headers.get(DIAZO_OFF_HEADER, "no")):
            return True

        path = request.path_info
        if self.ignored_pattern.search(path) is not None:
            return True

        return False

    def should_transform(self, response):
        """Determine if we should transform the response"""

        if asbool(response.headers.get(DIAZO_OFF_HEADER, "no")):
            return False

        content_type = response.headers.get("Content-Type")
        if not content_type or not (
            content_type.lower().startswith("text/html")
            or content_type.lower().startswith("application/xhtml+xml")
        ):
            return False

        content_encoding = response.headers.get("Content-Encoding")
        if content_encoding in (
            "zip",
            "deflate",
            "compress",
        ):
            return False

        status_code, reason = response.status.split(None, 1)
        if status_code.startswith("3") or status_code == "204" or status_code == "401":
            return False

        if response.content_length == 0:
            return False

        return True

    def reset_headers(self, response):
        # Remove any response headers that might change.
        response.content_range = None
        response.accept_ranges = None
        response.content_length = None
        response.content_md5 = None
        if self.remove_conditional_headers:
            response.last_modified = None
            response.etag = None
        # Set the output Content-Type
        response.content_type = self.content_type
        if self.content_type is not None:
            response.charset = self.charset


class DiazoMiddleware:
    """Invoke the Diazo transform as middleware"""

    def __init__(
        self,
        app,
        global_conf,
        rules,
        theme=None,
        prefix=None,
        includemode="document",
        debug=False,
        read_network=False,
        read_file=True,
        update_content_length=False,
        ignored_extensions=(
            "js",
            "css",
            "gif",
            "jpg",
            "jpeg",
            "pdf",
            "ps",
            "doc",
            "png",
            "ico",
            "mov",
            "mpg",
            "mpeg",
            "mp3",
            "m4a",
            "txt",
            "rtf",
            "swf",
            "wav",
            "zip",
            "wmv",
            "ppt",
            "gz",
            "tgz",
            "jar",
            "xls",
            "bmp",
            "tif",
            "tga",
            "hqx",
            "avi",
        ),
        environ_param_map=None,
        unquoted_params=None,
        doctype=None,
        content_type=None,
        filter_xpath=False,
        **params,  # NOQA: C816
    ):
        """Create the middleware. The parameters are:

        * ``rules``, the rules file
        * ``theme``, a URL to the theme file (may be a file:// URL)
        * ``debug``, set to True to recompile the theme on each request
        * ``prefix`` can be set to a string that will be prefixed to
          any *relative* URL referenced in an image, link or stylesheet in the
          theme HTML file before the theme is passed to the compiler. This
          allows a theme to be written so that it can be opened and views
          standalone on the filesystem, even if at runtime its static
          resources are going to be served from some other location. For
          example, an ``<img src="images/foo.jpg" />`` can be turned into
          ``<img src="/static/images/foo.jpg" />`` with a ``prefix`` of
          "/static".
        * ``includemode`` can be set to 'document', 'esi' or 'ssi' to change
          the way in which includes are processed
        * ``read_network``, should be set to True to allow resolving resources
          from the network.
        * ``read_file``, should be set to False to disallow resolving resources
          from the filesystem.
        * ``update_content_length``, can be set to True to update the
          Content-Length header when applying the transformation. When set to
          False (the default), the header is removed and it is left to the WSGI
          server recalculate or send a chunked response.
        * ``ignored_extensions`` can be set to a list of filename extensions
          for which the transformation should never be applied
        * ``environ_param_map`` can be set to a dict of environ keys to
          parameter names. The corresponding values will then be sent to the
          transformation as parameters.
        * ``unquoted_params``, can be set to a list of parameter names which
          will not be quoted.
        * ``doctype``, can be set to a string which will replace the default
          XHTML 1.0 transitional Doctype or that set in the Diazo theme. For
          example, "<!DOCTYPE html>".
        * ``content_type``, can be set to a string which will be set in the
          Content-Type header. By default it is inferred from the stylesheet.
        * ``charset``, can be set to a string which will be set in the
          Content-Type header. By default it is inferred from the stylesheet.
        * ``remove_conditional_headers``, should be set to True if the
        transformed output includes other files.
        * ``filter_xpath``, should be set to True to enable filter_xpath
          support for external includes.

        Additional keyword arguments will be passed to the theme
        transformation as parameters.
        """

        self.app = app
        self.global_conf = global_conf

        self.rules = rules
        self.theme = theme
        self.absolute_prefix = prefix
        self.includemode = includemode
        self.debug = asbool(debug)
        self.read_network = asbool(read_network)
        self.read_file = asbool(read_file)
        self.update_content_length = asbool(update_content_length)
        self.ignored_extensions = ignored_extensions
        self.doctype = doctype
        self.content_type = content_type
        self.unquoted_params = unquoted_params
        self.filter_xpath = asbool(filter_xpath)

        self.access_control = etree.XSLTAccessControl(
            read_file=self.read_file,
            write_file=False,
            create_dir=False,
            read_network=self.read_network,
            write_network=False,
        )
        self.transform_middleware = None
        self.filter_middleware = self.get_filter_middleware()

        self.environ_param_map = environ_param_map or {}
        self.environ_param_map.update(
            {
                "diazo.path": "path",
                "diazo.query_string": "query_string",
                "diazo.host": "host",
                "diazo.scheme": "scheme",
            }
        )

        self.params = params.copy()

    def compile_theme(self):
        """Compile the Diazo theme, returning an lxml tree (containing an XSLT
        document)
        """

        filesystem_resolver = FilesystemResolver(self.app)
        wsgi_resolver = WSGIResolver(self.app)
        python_resolver = PythonResolver()
        network_resolver = NetworkResolver()

        rules_parser = etree.XMLParser(recover=False)
        rules_parser.resolvers.add(wsgi_resolver)
        rules_parser.resolvers.add(filesystem_resolver)
        rules_parser.resolvers.add(python_resolver)
        if self.read_network:
            rules_parser.resolvers.add(network_resolver)

        theme_parser = etree.HTMLParser()
        theme_parser.resolvers.add(wsgi_resolver)
        theme_parser.resolvers.add(filesystem_resolver)
        theme_parser.resolvers.add(python_resolver)
        if self.read_network:
            theme_parser.resolvers.add(network_resolver)

        xsl_params = self.params.copy()
        for value in self.environ_param_map.values():
            if value not in xsl_params:
                xsl_params[value] = None

        return compile_theme(
            self.rules,
            theme=self.theme,
            absolute_prefix=self.absolute_prefix,
            includemode=self.includemode,
            access_control=self.access_control,
            read_network=self.read_network,
            parser=theme_parser,
            rules_parser=rules_parser,
            xsl_params=xsl_params,
        )

    def get_transform_middleware(self):
        return XSLTMiddleware(
            self.app,
            self.global_conf,
            tree=self.compile_theme(),
            read_network=self.read_network,
            read_file=self.read_file,
            update_content_length=self.update_content_length,
            ignored_extensions=self.ignored_extensions,
            environ_param_map=self.environ_param_map,
            doctype=self.doctype,
            content_type=self.content_type,
            unquoted_params=self.unquoted_params,
            **self.params,  # NOQA: C815,S101
        )

    def get_filter_middleware(self):
        tree = pkg_parse("filter_xhtml.xsl")
        return XSLTMiddleware(
            self.app,
            self.global_conf,
            tree=tree,
            read_network=False,
            read_file=False,
            update_content_length=self.update_content_length,
            ignored_extensions=self.ignored_extensions,
            environ_param_map={"diazo.filter_xpath": "xpath"},
            doctype="",
            content_type=self.content_type,
            unquoted_params=["xpath"],
        )

    def __call__(self, environ, start_response):
        if self.filter_xpath:
            filter_xpath = ";filter_xpath="
            query_string = environ.get("QUERY_STRING", "")
            if filter_xpath in query_string:
                environ["QUERY_STRING"], xpath = query_string.rsplit(
                    filter_xpath,
                    1,
                )
                environ["diazo.filter_xpath"] = unquote_plus(xpath)
                return self.filter_middleware(environ, start_response)

        transform_middleware = self.transform_middleware
        if transform_middleware is None or self.debug:
            transform_middleware = self.get_transform_middleware()
        if transform_middleware is not None and not self.debug:
            self.transform_middleware = transform_middleware

        # Set up variables, some of which are used as transform parameters
        request = Request(environ)

        environ["diazo.rules"] = self.rules
        environ["diazo.absolute_prefix"] = self.absolute_prefix
        environ["diazo.path"] = request.path
        environ["diazo.query_string"] = request.query_string
        environ["diazo.host"] = request.host
        environ["diazo.scheme"] = request.scheme

        return transform_middleware(environ, start_response)



#!/usr/bin/env python
"""\
Usage: %prog RULES

RULES is a file defining a set of diazo rules in css syntax, e.g:

<rules xmlns="http://namespaces.plone.org/diazo"
       xmlns:css="http://namespaces.plone.org/diazo/css">
       <copy css:content="#content-wrapper" css:theme="#page-content"/>
</rules>\
"""


from cssselect import GenericTranslator
from diazo import utils
from lxml import etree
from optparse import OptionParser

import logging
import sys


logger = logging.getLogger("diazo")
usage = __doc__


class LocationPathTranslator(GenericTranslator):
    def xpath_descendant_combinator(self, left, right):
        """right is a child, grand-child or further descendant of left"""
        return left.join("//", right)


_generic_translator = GenericTranslator()
_location_path_translator = LocationPathTranslator()


def convert_css_selectors(rules):
    """Convert css rules to xpath rules element tree in place"""
    # XXX: There is a :root pseudo-class  # NOQA: T000
    # http://www.w3.org/TR/css3-selectors/#root-pseudo
    # We may wish to add support to lxml.cssselect for it some day.
    for element in rules.xpath(
        "//@*[namespace-uri()='{nsp}']/..".format(nsp=utils.namespaces["css"]),
    ):
        tag_namespace = utils.namespace(element.tag)
        css_prefix = element.attrib.get(
            utils.fullname(
                utils.namespaces["css"],
                "prefix",
            ),
            None,
        )
        for name, value in element.attrib.items():
            if not name.startswith(
                "{%s}" % utils.namespaces["css"],  # NOQA: S001
            ):
                continue
            localname = utils.localname(name)
            if localname == "prefix":
                continue
            if not value:
                element.attrib[localname] = ""
                continue
            if (
                tag_namespace == utils.namespaces["diazo"]
                and localname
                in (
                    "content",
                    "content-children",
                    "if-content",
                    "if-not-content",
                )
                or (
                    tag_namespace == utils.namespaces["xsl"] and localname in ("match",)
                )
            ):
                prefix = css_prefix or "//"
                tr = _location_path_translator
            else:
                prefix = css_prefix or "descendant-or-self::"
                tr = _generic_translator
            element.attrib[localname] = tr.css_to_xpath(value, prefix=prefix)

    return rules


def main():
    """Called from console script"""
    parser = OptionParser(usage=usage)
    parser.add_option(
        "-o",
        "--output",
        metavar="output.html",
        help="Output filename (instead of stdout)",
        dest="output",
        default=sys.stdout,
    )
    parser.add_option(
        "-p",
        "--pretty-print",
        action="store_true",
        help="Pretty print output",
        dest="pretty_print",
        default=False,
    )
    (options, args) = parser.parse_args()

    if len(args) != 1:
        parser.error("Invalid number of arguments")
    rules = etree.parse(args[0])
    convert_css_selectors(rules)
    rules.write(options.output, pretty_print=options.pretty_print)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""\
Usage: %prog -x TRANSFORM CONTENT

  TRANSFORM is the compiled theme transform
  CONTENT is an html file.

Usage: %prog -r RULES [options] CONTENT
"""

from diazo.compiler import compile_theme
from diazo.utils import _createOptionParser
from diazo.utils import AC_READ_FILE
from diazo.utils import AC_READ_NET
from diazo.utils import quote_param
from diazo.utils import split_params
from lxml import etree

import diazo.runtrace
import logging
import os.path
import sys


logger = logging.getLogger("diazo")
usage = __doc__


class RunResolver(etree.Resolver):
    def __init__(self, directory):
        self.directory = directory

    def resolve(self, url, id, context):
        # libxml2 does not do this correctly on it's own with the HTMLParser
        # but it does work in Apache
        if "://" in url or url.startswith("/"):
            # It seems we must explicitly resolve the url here
            return self.resolve_filename(url, context)
        url = os.path.join(self.directory, url)
        return self.resolve_filename(url, context)


def main():
    """Called from console script"""
    op = _createOptionParser(usage=usage)
    op.add_option(
        "-x",
        "--xsl",
        metavar="transform.xsl",
        help="XSL transform",
        dest="xsl",
        default=None,
    )
    op.add_option(
        "--path",
        metavar="PATH",
        help="URI path",
        dest="path",
        default=None,
    )
    op.add_option(
        "--parameters",
        metavar="param1=val1,param2=val2",
        help="Set the values of arbitrary parameters",
        dest="parameters",
        default=None,
    )
    op.add_option(
        "--runtrace-xml",
        metavar="runtrace.xml",
        help="Write an xml format runtrace to file",
        dest="runtrace_xml",
        default=None,
    )
    op.add_option(
        "--runtrace-html",
        metavar="runtrace.html",
        help="Write an html format runtrace to file",
        dest="runtrace_html",
        default=None,
    )
    (options, args) = op.parse_args()

    if len(args) > 2:
        op.error("Wrong number of arguments.")
    elif len(args) == 2:
        if options.xsl or options.rules:
            op.error("Wrong number of arguments.")
        path, content = args
        if path.lower().endswith(".xsl"):
            options.xsl = path
        else:
            options.rules = path
    elif len(args) == 1:
        (content,) = args
    else:
        op.error("Wrong number of arguments.")
    if options.rules is None and options.xsl is None:
        op.error("Must supply either options or rules")

    if options.trace:
        logger.setLevel(logging.DEBUG)

    runtrace = False
    if options.runtrace_xml or options.runtrace_html:
        runtrace = True

    parser = etree.HTMLParser()
    parser.resolvers.add(RunResolver(os.path.dirname(content)))

    if options.xsl is not None:
        output_xslt = etree.parse(options.xsl)
    else:
        xsl_params = None
        if options.xsl_params:
            xsl_params = split_params(options.xsl_params)

        output_xslt = compile_theme(
            rules=options.rules,
            theme=options.theme,
            extra=options.extra,
            parser=parser,
            read_network=options.read_network,
            absolute_prefix=options.absolute_prefix,
            includemode=options.includemode,
            indent=options.pretty_print,
            xsl_params=xsl_params,
            runtrace=runtrace,
        )

    if content == "-":
        content = sys.stdin

    if options.read_network:
        access_control = AC_READ_NET
    else:
        access_control = AC_READ_FILE

    transform = etree.XSLT(output_xslt, access_control=access_control)
    content_doc = etree.parse(content, parser=parser)
    params = {}
    if options.path is not None:
        params["path"] = f"'{options.path}'"

    if options.parameters:
        for key, value in split_params(options.parameters).items():
            params[key] = quote_param(value)

    output_html = transform(content_doc, **params)
    if isinstance(options.output, str):
        out = open(options.output, "w")
    else:
        out = options.output
    out.write(str(output_html))

    if runtrace:
        runtrace_doc = diazo.runtrace.generate_runtrace(
            rules=options.rules,
            error_log=transform.error_log,
        )
        if options.runtrace_xml:
            if options.runtrace_xml == "-":
                out = sys.stdout
            else:
                out = open(options.runtrace_xml, "w")
            runtrace_doc.write(
                out,
                encoding="utf-8",
                pretty_print=options.pretty_print,
            )
        if options.runtrace_html:
            if options.runtrace_html == "-":
                out = sys.stdout
            else:
                out = open(options.runtrace_html, "w")
            out.write(str(diazo.runtrace.runtrace_to_html(runtrace_doc)))

    for msg in transform.error_log:
        if not msg.message.startswith("<runtrace "):
            logger.warn(msg)


if __name__ == "__main__":
    main()



from setuptools import find_packages
from setuptools import setup


extras_require = {
    "wsgi": [
        "repoze.xmliter",
        "WebOb",
    ],
    "test": [
        "repoze.xmliter",
        "WebOb",
    ],
}


readme = open("README.rst").read()
changes = open("CHANGES.rst").read()
long_desc = readme + "\n\n" + changes

setup(
    name="diazo",
    version="2.0.3",
    description="""Diazo implements a Deliverance like language using a pure
        XSLT engine. With Diazo, you "compile" your theme and ruleset in one
        step, then use a superfast/simple transform on each request thereafter.
        Alternatively, compile your theme during development, check it into
        version control, and not touch Diazo during deployment.""",
    keywords="web theming",
    long_description=long_desc,
    packages=find_packages("src"),
    package_dir={"": "src"},
    include_package_data=True,
    zip_safe=False,
    author="Paul Everitt, Laurence Rowe and Martin Aspeli.",
    author_email="laurence@lrowe.co.uk",
    url="http://diazo.org",
    license="New BSD",
    classifiers=[
        "Development Status :: 6 - Mature",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: WSGI",
        "Topic :: Internet :: WWW/HTTP :: WSGI :: Middleware",
        "Topic :: Text Processing :: Markup :: XML",
    ],
    python_requires=">=3.8",
    install_requires=[
        #    'setuptools',
        "lxml",
        "cssselect",
    ],
    extras_require=extras_require,
    entry_points="""
        [console_scripts]
        diazocompiler = diazo.compiler:main
        diazorun = diazo.run:main
        diazopreprocessor = diazo.rules:main

        [paste.filter_app_factory]
        xslt = diazo.wsgi:XSLTMiddleware
        main = diazo.wsgi:DiazoMiddleware
        """,
)



def alltests():
    from diazo.tests import test_diazo
    from diazo.tests import test_trace
    from diazo.tests import test_wsgi

    import unittest

    modules = [
        test_trace,
        test_wsgi,
        test_diazo,
    ]
    return unittest.TestSuite([module.test_suite() for module in modules])



import os.path
import sys
import unittest


if __name__ == "__main__":
    __file__ = sys.argv[0]


def _testfile(filename):
    return "/".join(
        ("file://",)
        + os.path.split(
            os.path.abspath(
                os.path.dirname(__file__),
            ),
        )
        + (
            "test_wsgi_files",
            filename,
        ),
    )


HTML = b"""\
<html>
    <body>
        <h1>Content title</h1>
        <div id="content">Content content</div>
    </body>
</html>
"""

HTML_ALTERNATIVE = b"""\
<html>
    <body>
        <h1>Content title</h1>
        <div id="content">Alternative content</div>
    </body>
</html>
"""

XSLT = b"""\
<xsl:stylesheet
    xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
    xmlns:xhtml="http://www.w3.org/1999/xhtml"
    version="1.0"
    exclude-result-prefixes="xhtml">
    <xsl:template match="/">
    <html>
        <head>
            <title>Transformed</title>
        </head>
        <body>
            <xsl:copy-of select="//div[@id='content']" />
        </body>
    </html>
    </xsl:template>
</xsl:stylesheet>
"""

XSLT_XHTML = b"""\
<xsl:stylesheet
    xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
    xmlns:xhtml="http://www.w3.org/1999/xhtml"
    version="1.0"
    exclude-result-prefixes="xhtml">
  <xsl:output method="xml" indent="no" omit-xml-declaration="yes"
    media-type="application/xhtml+xml" encoding="UTF-8"
    doctype-public="-//W3C//DTD XHTML 1.0 Transitional//EN"
    doctype-system="http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd"/>
  <xsl:template match="/">
    <html>
      <head>
        <title>Transformed</title>
      </head>
      <body>
        <xsl:copy-of select="//div[@id='content']" />
        <br/>
      </body>
    </html>
  </xsl:template>
</xsl:stylesheet>
"""

XSLT_HTML = b"""\
<xsl:stylesheet
    xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
    xmlns:xhtml="http://www.w3.org/1999/xhtml"
    version="1.0"
    exclude-result-prefixes="xhtml">
  <xsl:output method="html" indent="no" omit-xml-declaration="yes"
      media-type="text/html" encoding="UTF-8"
      doctype-public="-//W3C//DTD HTML 4.01//EN"
      doctype-system="http://www.w3.org/TR/html4/strict.dtd"/>
  <xsl:template match="/">
    <html>
        <head>
            <title>Transformed</title>
        </head>
        <body>
            <xsl:copy-of select="//div[@id='content']" />
            <br/>
        </body>
    </html>
  </xsl:template>
</xsl:stylesheet>
"""

# Note that this can only work with an html output method. Setting the doctype
# on the middleware along with the xml output method and an XHTML 1.0 doctype
# in the stylesheet is required for XHTML compatible output.
XSLT_HTML5 = b"""\
<xsl:stylesheet
    xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
    xmlns:xhtml="http://www.w3.org/1999/xhtml"
    version="1.0"
    exclude-result-prefixes="xhtml">
    <xsl:template match="/">
    <xsl:text
      disable-output-escaping="yes">&lt;!DOCTYPE html&gt;&#10;</xsl:text>
    <html>
        <head>
            <title>Transformed</title>
        </head>
        <body>
            <xsl:copy-of select="//div[@id='content']" />
            <br/>
        </body>
    </html>
    </xsl:template>
</xsl:stylesheet>
"""

XSLT_PARAM = b"""\
<xsl:stylesheet
    xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
    xmlns:xhtml="http://www.w3.org/1999/xhtml"
    version="1.0"
    exclude-result-prefixes="xhtml">
    <xsl:param name="someparam">defaultvalue</xsl:param>
    <xsl:template match="/">
    <html>
        <head>
            <title>Transformed</title>
        </head>
        <body>
            <xsl:copy-of select="//div[@id='content']" />
            <p><xsl:value-of select="$someparam" /></p>
        </body>
    </html>
    </xsl:template>
</xsl:stylesheet>
"""


class TestXSLTMiddleware(unittest.TestCase):
    def test_transform_filename(self):
        from diazo.wsgi import XSLTMiddleware
        from webob import Request

        import os
        import tempfile

        _, filename = tempfile.mkstemp()
        with open(filename, "wb") as fp:
            fp.write(XSLT)

        def application(environ, start_response):
            status = "200 OK"
            response_headers = [("Content-Type", "text/html")]
            start_response(status, response_headers)
            return [HTML]

        app = XSLTMiddleware(
            application,
            {},
            filename=filename,
        )
        os.unlink(filename)

        request = Request.blank("/")
        response = request.get_response(app)

        self.assertEqual(
            response.headers["Content-Type"],
            "text/html; charset=UTF-8",
        )
        self.assertTrue(
            b'<div id="content">Content content</div>' in response.body,
        )
        self.assertTrue(b"<title>Transformed</title>" in response.body)

    def test_transform_tree(self):
        from diazo.wsgi import XSLTMiddleware
        from lxml import etree
        from webob import Request

        def application(environ, start_response):
            status = "200 OK"
            response_headers = [("Content-Type", "text/html")]
            start_response(status, response_headers)
            return [HTML]

        app = XSLTMiddleware(
            application,
            {},
            tree=etree.fromstring(XSLT),
        )

        request = Request.blank("/")
        response = request.get_response(app)

        self.assertEqual(
            response.headers["Content-Type"],
            "text/html; charset=UTF-8",
        )
        self.assertTrue(
            b'<div id="content">Content content</div>' in response.body,
        )
        self.assertTrue(b"<title>Transformed</title>" in response.body)

    def test_head_request(self):
        from diazo.wsgi import XSLTMiddleware
        from lxml import etree
        from webob import Request

        def application(environ, start_response):
            status = "200 OK"
            response_headers = [
                ("Content-Type", "text/html"),
                ("Content-Length", str(len(HTML))),
            ]
            start_response(status, response_headers)
            return [""]  # Empty response for HEAD request

        app = XSLTMiddleware(
            application,
            {},
            tree=etree.fromstring(XSLT),
        )

        env = dict(REQUEST_METHOD="HEAD")
        request = Request.blank("/", environ=env)
        # The *real* test is whether or not an exception is raised here.
        response = request.get_response(app)

        # Response headers for HEAD request must be updated.
        self.assertEqual(response.headers.get("Content-Length"), None)
        self.assertFalse(response.body)

    def test_update_content_length(self):
        from diazo.wsgi import XSLTMiddleware
        from lxml import etree
        from webob import Request

        def application(environ, start_response):
            status = "200 OK"
            response_headers = [
                ("Content-Type", "text/html"),
                ("Content-Length", str(len(HTML))),
            ]
            start_response(status, response_headers)
            return [HTML]

        app = XSLTMiddleware(
            application,
            {},
            tree=etree.fromstring(XSLT),
            update_content_length=True,
        )

        request = Request.blank("/")
        response = request.get_response(app)

        self.assertEqual(response.headers["Content-Length"], "178")

    def test_dont_update_content_length(self):
        from diazo.wsgi import XSLTMiddleware
        from lxml import etree
        from webob import Request

        def application(environ, start_response):
            status = "200 OK"
            response_headers = [
                ("Content-Type", "text/html"),
                ("Content-Length", "1"),
            ]
            start_response(status, response_headers)
            return [HTML]

        app = XSLTMiddleware(
            application,
            {},
            tree=etree.fromstring(XSLT),
        )

        request = Request.blank("/")
        response = request.get_response(app)

        self.assertEqual(response.headers.get("Content-Length"), None)

    def test_content_length_zero(self):
        from diazo.wsgi import XSLTMiddleware
        from lxml import etree
        from webob import Request

        def application(environ, start_response):
            status = "200 OK"
            response_headers = [
                ("Content-Type", "text/html"),
                ("Content-Length", "0"),
            ]
            start_response(status, response_headers)
            return [""]

        app = XSLTMiddleware(
            application,
            {},
            tree=etree.fromstring(XSLT),
            update_content_length=True,
        )

        request = Request.blank("/")
        response = request.get_response(app)

        self.assertEqual(response.headers["Content-Length"], "0")

    def test_content_empty(self):
        from diazo.wsgi import XSLTMiddleware
        from lxml import etree
        from webob import Request

        def application(environ, start_response):
            status = "200 OK"
            response_headers = [
                ("Content-Type", "text/html"),
                ("Content-MD5", "d41d8cd98f00b204e9800998ecf8427e"),
            ]
            start_response(status, response_headers)
            return [b""]

        app = XSLTMiddleware(
            application,
            {},
            tree=etree.fromstring(XSLT),
            update_content_length=True,
        )

        request = Request.blank("/")
        response = request.get_response(app)

        self.assertEqual(
            response.headers["Content-MD5"],
            "d41d8cd98f00b204e9800998ecf8427e",
        )

    def test_content_range(self):
        from diazo.wsgi import XSLTMiddleware
        from lxml import etree
        from webob import Request

        def application(environ, start_response):
            status = "200 OK"
            content_length = len(HTML)
            content_range = "bytes {start:d}-{end:d}/{length:d}".format(
                start=0,
                end=content_length - 1,
                length=content_length,
            )
            response_headers = [
                ("Content-Type", "text/html"),
                ("Content-Range", content_range),
                ("Content-Length", str(content_length)),
            ]
            start_response(status, response_headers)
            return [HTML]

        app = XSLTMiddleware(
            application,
            {},
            tree=etree.fromstring(XSLT),
        )

        request = Request.blank("/")
        response = request.get_response(app)

        self.assertFalse("Content-Range" in response.headers)

    def test_no_content_length(self):
        from diazo.wsgi import XSLTMiddleware
        from lxml import etree
        from webob import Request

        def application(environ, start_response):
            status = "200 OK"
            response_headers = [("Content-Type", "text/html")]
            start_response(status, response_headers)
            return [HTML]

        app = XSLTMiddleware(
            application,
            {},
            tree=etree.fromstring(XSLT),
            set_content_length=False,
        )

        request = Request.blank("/")
        response = request.get_response(app)

        self.assertFalse("Content-Length" in response.headers)

    def test_doctype_html(self):
        from diazo.wsgi import XSLTMiddleware
        from lxml import etree
        from webob import Request

        def application(environ, start_response):
            status = "200 OK"
            response_headers = [("Content-Type", "text/html")]
            start_response(status, response_headers)
            return [HTML]

        app = XSLTMiddleware(
            application,
            {},
            tree=etree.fromstring(XSLT),
        )

        request = Request.blank("/")
        response = request.get_response(app)

        self.assertEqual(
            response.headers["Content-Type"],
            "text/html; charset=UTF-8",
        )

    def test_doctype_xhtml(self):
        from diazo.wsgi import XSLTMiddleware
        from lxml import etree
        from webob import Request

        def application(environ, start_response):
            status = "200 OK"
            response_headers = [("Content-Type", "text/html")]
            start_response(status, response_headers)
            return [HTML]

        app = XSLTMiddleware(
            application,
            {},
            tree=etree.fromstring(XSLT_XHTML),
        )

        request = Request.blank("/")
        response = request.get_response(app)

        self.assertEqual(
            response.headers["Content-Type"],
            "application/xhtml+xml; charset=UTF-8",
        )

    def test_doctype_html5(self):
        from diazo.wsgi import XSLTMiddleware
        from lxml import etree
        from webob import Request

        def application(environ, start_response):
            status = "200 OK"
            response_headers = [("Content-Type", "text/html")]
            start_response(status, response_headers)
            return [HTML]

        app = XSLTMiddleware(
            application,
            {},
            tree=etree.fromstring(XSLT_XHTML),
            doctype="<!DOCTYPE html>",
        )

        request = Request.blank("/")
        response = request.get_response(app)

        self.assertTrue(response.body.startswith(b"<!DOCTYPE html>\n<html"))

    def test_ignored_extension(self):
        from diazo.wsgi import XSLTMiddleware
        from lxml import etree
        from webob import Request

        def application(environ, start_response):
            status = "200 OK"
            response_headers = [("Content-Type", "text/html")]
            start_response(status, response_headers)
            return [HTML]

        app = XSLTMiddleware(
            application,
            {},
            tree=etree.fromstring(XSLT),
            ignored_extensions=("html",),
        )

        request = Request.blank("/index.html")
        response = request.get_response(app)

        self.assertEqual(response.body, HTML)

        request = Request.blank("/")
        response = request.get_response(app)

        self.assertTrue(
            b'<div id="content">Content content</div>' in response.body,
        )
        self.assertTrue(b"<title>Transformed</title>" in response.body)

    def test_diazo_off_request_header(self):
        from diazo.wsgi import XSLTMiddleware
        from lxml import etree
        from webob import Request

        def application(environ, start_response):
            status = "200 OK"
            response_headers = [("Content-Type", "text/html")]
            start_response(status, response_headers)
            return [HTML]

        app = XSLTMiddleware(
            application,
            {},
            tree=etree.fromstring(XSLT),
        )

        request = Request.blank("/")
        request.headers["X-Diazo-Off"] = "yes"
        response = request.get_response(app)

        self.assertEqual(response.body, HTML)

        request = Request.blank("/")
        request.headers["X-Diazo-Off"] = "no"
        response = request.get_response(app)

        self.assertTrue(
            b'<div id="content">Content content</div>' in response.body,
        )
        self.assertTrue(b"<title>Transformed</title>" in response.body)

    def test_diazo_off_response_header(self):
        from diazo.wsgi import XSLTMiddleware
        from lxml import etree
        from webob import Request

        def application1(environ, start_response):
            status = "200 OK"
            response_headers = [
                ("Content-Type", "text/html"),
                ("X-Diazo-Off", "yes"),
            ]
            start_response(status, response_headers)
            return [HTML]

        app = XSLTMiddleware(application1, {}, tree=etree.fromstring(XSLT))

        request = Request.blank("/")
        response = request.get_response(app)

        self.assertEqual(response.body, HTML)

        def application2(environ, start_response):
            status = "200 OK"
            response_headers = [
                ("Content-Type", "text/html"),
                ("X-Diazo-Off", "no"),
            ]
            start_response(status, response_headers)
            return [HTML]

        app = XSLTMiddleware(
            application2,
            {},
            tree=etree.fromstring(XSLT),
        )

        request = Request.blank("/")
        response = request.get_response(app)

        self.assertTrue(
            b'<div id="content">Content content</div>' in response.body,
        )
        self.assertTrue(b"<title>Transformed</title>" in response.body)

    def test_non_html_content_type(self):
        from diazo.wsgi import XSLTMiddleware
        from lxml import etree
        from webob import Request

        def application1(environ, start_response):
            status = "200 OK"
            response_headers = [("Content-Type", "text/plain")]
            start_response(status, response_headers)
            return [HTML]

        app = XSLTMiddleware(
            application1,
            {},
            tree=etree.fromstring(XSLT),
        )

        request = Request.blank("/")
        response = request.get_response(app)

        self.assertEqual(response.body, HTML)

        def application2(environ, start_response):
            status = "200 OK"
            response_headers = [("Content-Type", "text/html")]
            start_response(status, response_headers)
            return [HTML]

        app = XSLTMiddleware(
            application2,
            {},
            tree=etree.fromstring(XSLT),
        )

        request = Request.blank("/")
        response = request.get_response(app)

        self.assertTrue(
            b'<div id="content">Content content</div>' in response.body,
        )
        self.assertTrue(b"<title>Transformed</title>" in response.body)

    def test_content_encoding(self):
        from diazo.wsgi import XSLTMiddleware
        from lxml import etree
        from webob import Request

        def application1(environ, start_response):
            status = "200 OK"
            response_headers = [
                ("Content-Type", "text/html"),
                ("Content-Encoding", "zip"),
            ]
            start_response(status, response_headers)
            return [HTML]

        app = XSLTMiddleware(application1, {}, tree=etree.fromstring(XSLT))

        request = Request.blank("/")
        response = request.get_response(app)

        self.assertEqual(response.body, HTML)

        def application2(environ, start_response):
            status = "200 OK"
            response_headers = [("Content-Type", "text/html")]
            start_response(status, response_headers)
            return [HTML]

        app = XSLTMiddleware(
            application2,
            {},
            tree=etree.fromstring(XSLT),
        )

        request = Request.blank("/")
        response = request.get_response(app)

        self.assertTrue(
            b'<div id="content">Content content</div>' in response.body,
        )
        self.assertTrue(b"<title>Transformed</title>" in response.body)

    def test_301(self):
        from diazo.wsgi import XSLTMiddleware
        from lxml import etree
        from webob import Request

        def application1(environ, start_response):
            status = "301 MOVED PERMANENTLY"
            response_headers = [("Content-Type", "text/html")]
            start_response(status, response_headers)
            return [HTML]

        app = XSLTMiddleware(
            application1,
            {},
            tree=etree.fromstring(XSLT),
        )

        request = Request.blank("/")
        response = request.get_response(app)

        self.assertEqual(response.body, HTML)

        def application2(environ, start_response):
            status = "200 OK"
            response_headers = [("Content-Type", "text/html")]
            start_response(status, response_headers)
            return [HTML]

        app = XSLTMiddleware(
            application2,
            {},
            tree=etree.fromstring(XSLT),
        )

        request = Request.blank("/")
        response = request.get_response(app)

        self.assertTrue(
            b'<div id="content">Content content</div>' in response.body,
        )
        self.assertTrue(b"<title>Transformed</title>" in response.body)

    def test_302(self):
        from diazo.wsgi import XSLTMiddleware
        from lxml import etree
        from webob import Request

        def application1(environ, start_response):
            status = "302 MOVED"
            response_headers = [("Content-Type", "text/html")]
            start_response(status, response_headers)
            return [HTML]

        app = XSLTMiddleware(
            application1,
            {},
            tree=etree.fromstring(XSLT),
        )

        request = Request.blank("/")
        response = request.get_response(app)

        self.assertEqual(response.body, HTML)

        def application2(environ, start_response):
            status = "200 OK"
            response_headers = [("Content-Type", "text/html")]
            start_response(status, response_headers)
            return [HTML]

        app = XSLTMiddleware(
            application2,
            {},
            tree=etree.fromstring(XSLT),
        )

        request = Request.blank("/")
        response = request.get_response(app)

        self.assertTrue(
            b'<div id="content">Content content</div>' in response.body,
        )
        self.assertTrue(b"<title>Transformed</title>" in response.body)

    def test_304(self):
        from diazo.wsgi import XSLTMiddleware
        from lxml import etree
        from webob import Request

        def application1(environ, start_response):
            status = "304 NOT MODIFIED"
            response_headers = [("Content-Type", "text/html")]
            start_response(status, response_headers)
            return [HTML]

        app = XSLTMiddleware(application1, {}, tree=etree.fromstring(XSLT))

        request = Request.blank("/")
        response = request.get_response(app)

        self.assertEqual(response.body, HTML)

        def application2(environ, start_response):
            status = "200 OK"
            response_headers = [("Content-Type", "text/html")]
            start_response(status, response_headers)
            return [HTML]

        app = XSLTMiddleware(application2, {}, tree=etree.fromstring(XSLT))

        request = Request.blank("/")
        response = request.get_response(app)

        self.assertTrue(
            b'<div id="content">Content content</div>' in response.body,
        )
        self.assertTrue(b"<title>Transformed</title>" in response.body)

    def test_204(self):
        from diazo.wsgi import XSLTMiddleware
        from lxml import etree
        from webob import Request

        def application1(environ, start_response):
            status = "204 NO CONTENT"
            response_headers = [("Content-Type", "text/html")]
            start_response(status, response_headers)
            return [HTML]

        app = XSLTMiddleware(application1, {}, tree=etree.fromstring(XSLT))

        request = Request.blank("/")
        response = request.get_response(app)

        self.assertEqual(response.body, HTML)

        def application2(environ, start_response):
            status = "200 OK"
            response_headers = [("Content-Type", "text/html")]
            start_response(status, response_headers)
            return [HTML]

        app = XSLTMiddleware(application2, {}, tree=etree.fromstring(XSLT))

        request = Request.blank("/")
        response = request.get_response(app)

        self.assertTrue(
            b'<div id="content">Content content</div>' in response.body,
        )
        self.assertTrue(b"<title>Transformed</title>" in response.body)

    def test_401(self):
        from diazo.wsgi import XSLTMiddleware
        from lxml import etree
        from webob import Request

        def application1(environ, start_response):
            status = "401 UNAUTHORIZED"
            response_headers = [("Content-Type", "text/html")]
            start_response(status, response_headers)
            return [HTML]

        app = XSLTMiddleware(
            application1,
            {},
            tree=etree.fromstring(XSLT),
        )

        request = Request.blank("/")
        response = request.get_response(app)

        self.assertEqual(response.body, HTML)

        def application2(environ, start_response):
            status = "200 OK"
            response_headers = [("Content-Type", "text/html")]
            start_response(status, response_headers)
            return [HTML]

        app = XSLTMiddleware(
            application2,
            {},
            tree=etree.fromstring(XSLT),
        )

        request = Request.blank("/")
        response = request.get_response(app)

        self.assertTrue(
            b'<div id="content">Content content</div>' in response.body,
        )
        self.assertTrue(b"<title>Transformed</title>" in response.body)

    def test_html_serialization(self):
        from diazo.wsgi import XSLTMiddleware
        from lxml import etree
        from webob import Request

        def application(environ, start_response):
            status = "200 OK"
            response_headers = [("Content-Type", "text/html")]
            start_response(status, response_headers)
            return [HTML]

        app = XSLTMiddleware(application, {}, tree=etree.fromstring(XSLT_HTML))
        request = Request.blank("/")
        response = request.get_response(app)

        # HTML serialisation
        self.assertTrue(
            b'<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01//EN" '
            b'"http://www.w3.org/TR/html4/strict.dtd">' in response.body,
        )
        self.assertTrue(b"<br>" in response.body)

        app = XSLTMiddleware(
            application,
            {},
            tree=etree.fromstring(XSLT_XHTML),
        )
        request = Request.blank("/")
        response = request.get_response(app)

        # XHTML serialisation
        self.assertTrue(
            b'<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" '
            b'"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">'
            in response.body,
        )
        self.assertTrue(b"<br />" in response.body)

        app = XSLTMiddleware(
            application,
            {},
            tree=etree.fromstring(XSLT_HTML5),
        )
        request = Request.blank("/")
        response = request.get_response(app)

        # HTML 5 serialisation
        self.assertTrue(b"<!DOCTYPE html>" in response.body)
        self.assertTrue(b"<br/>" in response.body)

    def test_environ_param(self):
        from diazo.wsgi import XSLTMiddleware
        from lxml import etree
        from webob import Request

        def application(environ, start_response):
            status = "200 OK"
            response_headers = [("Content-Type", "text/html")]
            start_response(status, response_headers)
            return [HTML]

        app = XSLTMiddleware(
            application,
            {},
            tree=etree.fromstring(XSLT_PARAM),
            environ_param_map={"test.param1": "someparam"},
        )

        request = Request.blank("/")
        response = request.get_response(app)

        self.assertTrue(b"<p>defaultvalue</p>" in response.body)

        request = Request.blank("/")
        request.environ["test.param1"] = "value1"
        response = request.get_response(app)

        self.assertTrue(b"<p>value1</p>" in response.body)

    def test_params(self):
        from diazo.wsgi import XSLTMiddleware
        from lxml import etree
        from webob import Request

        def application(environ, start_response):
            status = "200 OK"
            response_headers = [("Content-Type", "text/html")]
            start_response(status, response_headers)
            return [HTML]

        app = XSLTMiddleware(
            application,
            {},
            tree=etree.fromstring(XSLT_PARAM),
        )
        request = Request.blank("/")
        response = request.get_response(app)

        self.assertTrue(b"<p>defaultvalue</p>" in response.body)

        app = XSLTMiddleware(
            application,
            {},
            tree=etree.fromstring(XSLT_PARAM),
            someparam="value1",
        )
        request = Request.blank("/")
        response = request.get_response(app)

        self.assertTrue(b"<p>value1</p>" in response.body)


class TestDiazoMiddleware(unittest.TestCase):
    def test_simple_transform(self):
        from diazo.wsgi import DiazoMiddleware
        from webob import Request

        def application(environ, start_response):
            status = "200 OK"
            response_headers = [("Content-Type", "text/html")]
            start_response(status, response_headers)
            return [HTML]

        app = DiazoMiddleware(
            application,
            {},
            _testfile("simple_transform.xml"),
        )
        request = Request.blank("/")
        response = request.get_response(app)

        self.assertTrue(
            b'<div id="content">Content content</div>' in response.body,
        )
        self.assertFalse(
            b'<div id="content">Theme content</div>' in response.body,
        )
        self.assertTrue(b"<title>Transformed</title>" in response.body)

    def test_doctype_html5(self):
        from diazo.wsgi import DiazoMiddleware
        from webob import Request

        def application(environ, start_response):
            status = "200 OK"
            response_headers = [("Content-Type", "text/html")]
            start_response(status, response_headers)
            return [HTML]

        app = DiazoMiddleware(
            application,
            {},
            _testfile("simple_transform.xml"),
            doctype="<!DOCTYPE html>",
        )
        request = Request.blank("/")
        response = request.get_response(app)

        self.assertTrue(response.body.startswith(b"<!DOCTYPE html>\n<html"))

    def test_with_theme(self):
        from diazo.wsgi import DiazoMiddleware
        from webob import Request

        def application(environ, start_response):
            status = "200 OK"
            response_headers = [("Content-Type", "text/html")]
            start_response(status, response_headers)
            return [HTML]

        app = DiazoMiddleware(
            application,
            {},
            _testfile("explicit_theme.xml"),
            theme=_testfile("theme.html"),
        )
        request = Request.blank("/")
        response = request.get_response(app)

        self.assertTrue(
            b'<div id="content">Content content</div>' in response.body,
        )
        self.assertFalse(
            b'<div id="content">Theme content</div>' in response.body,
        )
        self.assertTrue(b"<title>Transformed</title>" in response.body)

    def test_absolute_prefix(self):
        from diazo.wsgi import DiazoMiddleware
        from webob import Request

        def application(environ, start_response):
            status = "200 OK"
            response_headers = [("Content-Type", "text/html")]
            start_response(status, response_headers)
            return [HTML]

        app = DiazoMiddleware(
            application,
            {},
            _testfile("simple_transform.xml"),
        )
        request = Request.blank("/")
        response = request.get_response(app)

        self.assertTrue(
            b'<div id="content">Content content</div>' in response.body,
        )
        self.assertFalse(
            b'<div id="content">Theme content</div>' in response.body,
        )
        self.assertTrue(b"<title>Transformed</title>" in response.body)
        self.assertTrue(
            b'<link rel="stylesheet" href="./theme.css" />' in response.body,
        )

        app = DiazoMiddleware(
            application,
            {},
            _testfile("simple_transform.xml"),
            prefix="/static",
        )
        request = Request.blank("/")
        response = request.get_response(app)

        self.assertTrue(
            b'<div id="content">Content content</div>' in response.body,
        )
        self.assertFalse(
            b'<div id="content">Theme content</div>' in response.body,
        )
        self.assertTrue(b"<title>Transformed</title>" in response.body)
        self.assertTrue(
            b'<link rel="stylesheet" href="/static/theme.css" />' in response.body,
        )

    def test_path_param(self):
        from diazo.wsgi import DiazoMiddleware
        from webob import Request

        def application(environ, start_response):
            status = "200 OK"
            response_headers = [("Content-Type", "text/html")]
            start_response(status, response_headers)
            return [HTML]

        app = DiazoMiddleware(application, {}, _testfile("path_param.xml"))
        request = Request.blank("/")
        response = request.get_response(app)

        self.assertFalse(
            b'<div id="content">Content content</div>' in response.body,
        )
        self.assertTrue(
            b'<div id="content">Theme content</div>' in response.body,
        )
        self.assertTrue(b"<title>Transformed</title>" in response.body)

        request = Request.blank("/index.html")
        response = request.get_response(app)

        self.assertTrue(
            b'<div id="content">Content content</div>' in response.body,
        )
        self.assertFalse(
            b'<div id="content">Theme content</div>' in response.body,
        )
        self.assertTrue(b"<title>Transformed</title>" in response.body)

    def test_custom_environ_param(self):
        from diazo.wsgi import DiazoMiddleware
        from webob import Request

        def application(environ, start_response):
            status = "200 OK"
            response_headers = [("Content-Type", "text/html")]
            start_response(status, response_headers)
            return [HTML]

        app = DiazoMiddleware(
            application,
            {},
            _testfile("custom_param.xml"),
            environ_param_map={"test.param1": "someparam"},
        )

        request = Request.blank("/")
        response = request.get_response(app)

        self.assertFalse(
            b'<div id="content">Content content</div>' in response.body,
        )
        self.assertTrue(
            b'<div id="content">Theme content</div>' in response.body,
        )
        self.assertTrue(b"<title>Transformed</title>" in response.body)

        request = Request.blank("/")
        request.environ["test.param1"] = "value1"
        response = request.get_response(app)

        self.assertTrue(
            b'<div id="content">Content content</div>' in response.body,
        )
        self.assertFalse(
            b'<div id="content">Theme content</div>' in response.body,
        )
        self.assertTrue(b"<title>Transformed</title>" in response.body)

        request = Request.blank("/")
        request.environ["test.param1"] = "value2"
        response = request.get_response(app)

        self.assertFalse(
            b'<div id="content">Content content</div>' in response.body,
        )
        self.assertTrue(
            b'<div id="content">Theme content</div>' in response.body,
        )
        self.assertTrue(b"<title>Transformed</title>" in response.body)

    def test_custom_param(self):
        from diazo.wsgi import DiazoMiddleware
        from webob import Request

        def application(environ, start_response):
            status = "200 OK"
            response_headers = [("Content-Type", "text/html")]
            start_response(status, response_headers)
            return [HTML]

        app = DiazoMiddleware(
            application,
            {},
            _testfile("custom_param.xml"),
            someparam="value1",
        )
        request = Request.blank("/")
        response = request.get_response(app)

        self.assertTrue(
            b'<div id="content">Content content</div>' in response.body,
        )
        self.assertFalse(
            b'<div id="content">Theme content</div>' in response.body,
        )
        self.assertTrue(b"<title>Transformed</title>" in response.body)

        app = DiazoMiddleware(
            application,
            {},
            _testfile("custom_param.xml"),
            someparam="value2",
        )
        request = Request.blank("/")
        response = request.get_response(app)

        self.assertFalse(
            b'<div id="content">Content content</div>' in response.body,
        )
        self.assertTrue(
            b'<div id="content">Theme content</div>' in response.body,
        )
        self.assertTrue(b"<title>Transformed</title>" in response.body)

    def test_subrequest(self):
        from diazo.wsgi import DiazoMiddleware
        from webob import Request

        def application(environ, start_response):
            status = "200 OK"
            response_headers = [("Content-Type", "text/html")]
            start_response(status, response_headers)

            request = Request(environ)
            if request.path.endswith("/other.html"):
                return [HTML_ALTERNATIVE]
            else:
                return [HTML]

        app = DiazoMiddleware(application, {}, _testfile("subrequest.xml"))
        request = Request.blank("/")
        response = request.get_response(app)

        self.assertTrue(
            b'<div id="content">Alternative content</div>' in response.body,
        )
        self.assertFalse(
            b'<div id="content">Theme content</div>' in response.body,
        )
        self.assertTrue(b"<title>Transformed</title>" in response.body)

    def test_esi(self):
        from diazo.wsgi import DiazoMiddleware
        from webob import Request

        def application(environ, start_response):
            status = "200 OK"
            response_headers = [("Content-Type", "text/html")]
            start_response(status, response_headers)

            request = Request(environ)
            if request.path.endswith("/other.html"):
                return [HTML_ALTERNATIVE]
            else:
                return [HTML]

        app = DiazoMiddleware(
            application,
            {},
            _testfile("esi.xml"),
            filter_xpath=True,
        )
        request = Request.blank("/")
        response = request.get_response(app)

        self.assertTrue(
            b"""<esi:include src="/other.html?;"""
            b"""filter_xpath=//*[@id%20=%20'content']">"""
            b"""</esi:include>""" in response.body,
        )
        self.assertFalse(
            b'<div id="content">Theme content</div>' in response.body,
        )
        self.assertTrue(b"<title>Transformed</title>" in response.body)

        request = Request.blank(
            """/other.html?;filter_xpath=//*[@id%20=%20'content']""",
        )
        response = request.get_response(app)
        # Strip response body in this test due too
        # https://bugzilla.gnome.org/show_bug.cgi?id=652766
        self.assertEqual(
            b'<div id="content">Alternative content</div>',
            response.body.strip(),
        )


def test_suite():
    return unittest.defaultTestLoader.loadTestsFromName(__name__)



from diazo.utils import quote_param
from importlib.resources import files
from io import BytesIO
from io import StringIO
from lxml import etree

import configparser
import diazo.compiler
import diazo.run
import difflib
import os
import sys
import unittest


#
# Simple test runner for validating different diazo scenarios
#

if __name__ == "__main__":
    __file__ = sys.argv[0]


defaultsfn = str(files("diazo") / "tests" / "default-options.cfg")


def text_compare(t1, t2):
    # Copied from formencode.doctest_xml_compare.text_compare 2.0.1.
    # See note in xml_compare below.
    if not t1 and not t2:
        return True
    if t1 == "*" or t2 == "*":
        return True
    return (t1 or "").strip() == (t2 or "").strip()


def xml_compare(x1, x2):
    """Compare two xml items.

    Copied from formencode.doctest_xml_compare.xml_compare 2.0.1,
    without the (unused by us) optional 'reporter' argument.

    License: MIT

    Copyright (c) 2015 Ian Bicking and FormEncode Contributors

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
    THE SOFTWARE.
    """
    if x1.tag != x2.tag:
        return False
    for name, value in x1.attrib.items():
        if x2.attrib.get(name) != value:
            return False
    for name in x2.attrib:
        if name not in x1.attrib:
            return False
    if not text_compare(x1.text, x2.text):
        return False
    if not text_compare(x1.tail, x2.tail):
        return False
    cl1 = list(x1)
    cl2 = list(x2)
    if len(cl1) != len(cl2):
        return False
    i = 0
    for c1, c2 in zip(cl1, cl2):
        i += 1
        if not xml_compare(c1, c2):
            return False
    return True


class DiazoTestCase(unittest.TestCase):
    writefiles = os.environ.get("DiazoTESTS_WRITE_FILES", False)
    warnings = os.environ.get(
        "DiazoTESTS_WARN",
        "1",
    ).lower() not in ("0", "false", "off")

    testdir = os.path.realpath(__file__)

    @classmethod
    def suiteForParent(cls, parent, prefix):
        """Return a suite of diazo tests, one for each directory in parent."""
        suite = unittest.TestSuite()
        for name in os.listdir(parent):
            if name.startswith("."):
                continue
            path = os.path.join(parent, name)
            if not os.path.isdir(path):
                continue

            contentpath = os.path.join(path, "content.html")
            if not os.path.isfile(contentpath):
                continue

            test_cls = type(
                "{prefix:s}-{name:s}".format(
                    prefix=prefix,
                    name=name,
                ),
                (DiazoTestCase,),
                dict(testdir=path),
            )
            suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(test_cls))
        return suite

    def testAll(self):
        self.errors = BytesIO()
        config = configparser.ConfigParser()
        config.read([defaultsfn, os.path.join(self.testdir, "options.cfg")])

        themefn = None
        if config.get("diazotest", "theme"):
            themefn = os.path.join(
                self.testdir,
                config.get("diazotest", "theme"),
            )
        contentfn = os.path.join(self.testdir, "content.html")
        rulesfn = os.path.join(self.testdir, "rules.xml")
        xpathsfn = os.path.join(self.testdir, "xpaths.txt")
        xslfn = os.path.join(self.testdir, "compiled.xsl")
        outputfn = os.path.join(self.testdir, "output.html")

        xsl_params = {}
        extra_params = config.get("diazotest", "extra-params")
        if extra_params:
            for token in extra_params.split(" "):
                token_split = token.split(":")
                xsl_params[token_split[0]] = (
                    len(token_split) > 1 and token_split[1] or None
                )

        if not os.path.exists(rulesfn):
            return

        contentdoc = etree.parse(
            source=contentfn,
            base_url=contentfn,
            parser=etree.HTMLParser(),
        )

        # Make a compiled version
        theme_parser = etree.HTMLParser()
        ct = diazo.compiler.compile_theme(
            rules=rulesfn,
            theme=themefn,
            parser=theme_parser,
            absolute_prefix=config.get("diazotest", "absolute-prefix"),
            indent=config.getboolean("diazotest", "pretty-print"),
            xsl_params=xsl_params,
        )

        # Serialize / parse the theme - this can catch problems with escaping.
        cts = etree.tostring(ct, encoding="unicode")
        parser = etree.XMLParser()
        etree.fromstring(cts, parser=parser)

        # Compare to previous version
        if os.path.exists(xslfn):
            with open(xslfn) as f:
                old = f.read()
            new = cts
            if old != new:
                if self.writefiles:
                    with open(xslfn + ".old", "w") as f:
                        f.write(old)
                if self.warnings:
                    print("WARNING:", "compiled.xsl has CHANGED")
                    for line in difflib.unified_diff(
                        old.split("\n"),
                        new.split("\n"),
                        xslfn,
                        "now",
                    ):
                        print(line)

        # Write the compiled xsl out to catch unexpected changes
        if self.writefiles:
            with open(xslfn, "w") as f:
                f.write(cts)

        # Apply the compiled version, then test against desired output
        theme_parser.resolvers.add(diazo.run.RunResolver(self.testdir))
        processor = etree.XSLT(ct)
        params = {}
        params["path"] = "'{path:s}'".format(
            path=config.get("diazotest", "path"),
        )

        for key in xsl_params:
            try:
                params[key] = quote_param(config.get("diazotest", key))
            except configparser.NoOptionError:
                pass

        result = processor(contentdoc, **params)

        # Read the whole thing to strip off xhtml namespace.
        # If we had xslt 2.0 then we could use xpath-default-namespace.
        self.themed_string = str(result)
        self.themed_content = etree.ElementTree(
            file=StringIO(self.themed_string),
            parser=etree.HTMLParser(),
        )

        # remove the extra meta content type

        metas = self.themed_content.xpath("/html/head/meta[@http-equiv='Content-Type']")
        if metas:
            meta = metas[0]
            meta.getparent().remove(meta)

        if os.path.exists(xpathsfn):
            with open(xpathsfn) as f:
                for xpath in f.readlines():
                    # Read the XPaths from the file, skipping blank lines and
                    # comments
                    this_xpath = xpath.strip()
                    if not this_xpath or this_xpath[0] == "#":
                        continue
                    assert self.themed_content.xpath(
                        this_xpath
                    ), "{key:s}: {value:s}".format(  # NOQA: E501
                        key=xpathsfn,
                        value=this_xpath,
                    )

        # Compare to previous version
        if os.path.exists(outputfn):
            with open(outputfn, "rb") as f:
                old = f.read()
            new = self.themed_string
            if not xml_compare(
                etree.fromstring(old.strip()),
                etree.fromstring(new.strip()),
            ):
                # if self.writefiles:
                #    open(outputfn + '.old', 'w').write(old)
                for line in difflib.unified_diff(
                    old.split("\n"),
                    new.split("\n"),
                    outputfn,
                    "now",
                ):
                    print(line)
                assert old == new, "output.html has CHANGED"

        # Write out the result to catch unexpected changes
        if self.writefiles:
            with open(outputfn, "w") as f:
                f.write(self.themed_string)


def test_suite():
    suite = unittest.TestSuite()
    tests_dir = os.path.dirname(__file__)
    suite.addTest(DiazoTestCase.suiteForParent(tests_dir, "Test"))
    recipes_dir = os.path.join(
        tests_dir,
        "../../..",
        "docs/recipes",
    )
    if os.path.exists(os.path.join(recipes_dir, "diazo-tests-marker.txt")):
        # Could still be a 'System' package.
        suite.addTest(DiazoTestCase.suiteForParent(recipes_dir, "Recipe"))
    return suite
